import logging
import sys
import uuid
from datetime import timedelta, datetime as py_datetime
from django.core.cache import cache
from cached_property import cached_property
from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import AbstractBaseUser, BaseUserManager, PermissionsMixin, Group
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.db import models
from django.utils.crypto import salted_hmac
from graphql import ResolveInfo
import core


# from core.utils import validate_password
from django.contrib.auth.password_validation import validate_password
#from core.datetimes.ad_datetime import datetime as py_datetime
from django.conf import settings

from ..utils import filter_validity
from .base import *
from .versioned_model import *

logger = logging.getLogger(__name__)
from rest_framework import exceptions


class UserManager(BaseUserManager):

    def _create_core_user(self, **fields):
        user = User(**fields)
        user.save()
        return user

    def _create_tech_user(self, username, email, password, **extra_fields):
        tech = TechnicalUser(username=username, email=email, **extra_fields)
        tech.set_password(password)
        tech.save()
        return tech

    def create_user(self, username, password, email=None, **extra_fields):
        extra_fields.setdefault('is_staff', False)
        extra_fields['is_superuser'] = False
        self._create_tech_user(username, email, password, **extra_fields)

    def create_superuser(self, username, password=None, email=None, **extra_fields):
        extra_fields['is_staff'] = True
        extra_fields['is_superuser'] = True
        self._create_tech_user(username, email, password, **extra_fields)

    def auto_provision_user(self, **kwargs):
        # only auto-provision django user if registered as interactive user
        try:
            i_user = InteractiveUser.objects.get(
                login_name=kwargs['username'],
                *filter_validity())
        except InteractiveUser.DoesNotExist as e:
            raise exceptions.AuthenticationFailed("INCORRECT_CREDENTIALS") from e
        user = self._create_core_user(**kwargs)
        user.i_user = i_user
        user.save()
        if core.auto_provisioning_user_group:
            group = Group.objects.filter(
                name=core.auto_provisioning_user_group).first()
            if group:
                user_group = UserGroup(user=user, group=group)
                user_group.save()
            else:
                logger.error(f"Group {core.auto_provisioning_user_group} was not found")
        return user, True

    def get_or_create(self, **kwargs):
        user = User.objects.filter(username__iexact=kwargs.get("username")).first()
        if user:
            return user, False
        else:
            return self.auto_provision_user(**kwargs)


class TechnicalUser(AbstractBaseUser):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    username = models.CharField(max_length=50, unique=True)
    email = models.EmailField(blank=True, null=True)
    language = 'en'
    is_staff = models.BooleanField(default=False)
    is_superuser = models.BooleanField(default=False)
    validity_from = models.DateTimeField(blank=True, null=True, default = py_datetime.now)
    validity_to = models.DateTimeField(blank=True, null=True)
    is_imis_admin = False
    @property
    def id_for_audit(self):
        return -1

    USERNAME_FIELD = 'username'
    REQUIRED_FIELDS = ['password']

    def _bind_User(self):
        save_required = False
        try:
            usr = User.objects.get(t_user=self)
        except ObjectDoesNotExist:
            usr = User(username=self.username)
            usr.t_user = self
            save_required = True
        if usr.username != self.username:
            usr.username = self.username
            save_required = True
        if save_required:
            usr.shallow_save()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self._bind_User()

    class Meta:
        managed = True
        db_table = 'core_TechnicalUser'


class Role(VersionedModel):
    id = models.AutoField(db_column='RoleID', primary_key=True)
    uuid = models.CharField(db_column='RoleUUID', max_length=36, default=uuid.uuid4, unique=True)
    name = models.CharField(db_column='RoleName', max_length=50)
    alt_language = models.CharField(
        db_column='AltLanguage', max_length=50, blank=True, null=True)
    is_system = models.IntegerField(db_column='IsSystem')
    is_blocked = models.BooleanField(db_column='IsBlocked')
    audit_user_id = models.IntegerField(
        db_column='AuditUserID', blank=True, null=True)

    def natural_key(self):
        return (self.uuid,)

    @classmethod
    def get_queryset(cls, queryset, user):
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            pass
        return queryset

    class Meta:
        managed = True
        db_table = 'tblRole'


class RoleRight(VersionedModel):
    id = models.AutoField(db_column='RoleRightID', primary_key=True)
    role = models.ForeignKey(Role, models.DO_NOTHING,
                             db_column='RoleID', related_name="rights")
    right_id = models.IntegerField(db_column='RightID')
    audit_user_id = models.IntegerField(
        db_column='AuditUserId', blank=True, null=True)

    @classmethod
    def get_queryset(cls, queryset, user):
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            pass
        return queryset

    @classmethod
    def _get_by_uuid(cls, uuid_value):
        """Custom method to look up Role by UUID, which will be used when importing the fixture."""
        try:
            return Role.objects.get(uuid=uuid_value)
        except ObjectDoesNotExist:
            raise ValueError(f"Role with UUID {uuid_value} does not exist")

    class Meta:
        managed = True
        db_table = 'tblRoleRight'


class InteractiveUser(VersionedModel):
    id = models.AutoField(db_column="UserID", primary_key=True)
    uuid = models.CharField(db_column="UserUUID", max_length=36, default=uuid.uuid4, unique=True)
    language = models.ForeignKey(Language, models.DO_NOTHING, db_column="LanguageID")
    last_name = models.CharField(db_column="LastName", max_length=100)
    other_names = models.CharField(db_column="OtherNames", max_length=100)
    phone = models.CharField(db_column="Phone", max_length=50, blank=True, null=True)
    login_name = models.CharField(db_column="LoginName", max_length=50)
    last_login = models.DateTimeField(db_column="LastLogin", null=True, blank=True)
    health_facility_id = models.IntegerField(db_column="HFID", blank=True, null=True)

    audit_user_id = models.IntegerField(db_column="AuditUserID")
    # dummy_pwd is always blank. It is actually a transient field used in the Legacy to pass the clear text password in
    # a User object from the ASPX to the DAL where it is processed into/against password and private key/salt)
    # dummy_pwd = models.CharField(db_column='DummyPwd', max_length=25, blank=True, null=True)
    email = models.CharField(db_column="EmailId", max_length=200, blank=True, null=True)
    private_key = models.CharField(
        db_column="PrivateKey",
        max_length=256,
        blank=True,
        null=True,
        help_text="The private key is actually a password salt",
    )
    password = models.CharField(
        db_column="StoredPassword",
        max_length=256,
        blank=True,
        null=True,
        help_text="By default a SHA256 of the private key (salt) and password",
    )

    password_validity = models.DateTimeField(
        db_column="PasswordValidity", blank=True, null=True
    )
    is_associated = models.BooleanField(
        db_column="IsAssociated",
        blank=True,
        null=True,
        help_text="has a claim admin or enrolment officer account",
    )
    # deprecated
    role_id = models.IntegerField(db_column="RoleID", null=True, blank=True)

    @property
    def id_for_audit(self):
        return id

    @property
    def username(self):
        return self.login_name

    def get_username(self):
        return self.login_name

    @property
    def stored_password(self):
        return self.password

    @property
    def user(self):
        return self.user_set.first()

    @stored_password.setter
    def stored_password(self, value):
        logger.warn(
            "You should not use this property to set a password. Use 'password' instead."
        )
        self.password = value

    @property
    def is_staff(self):
        return self.is_superuser


    @property
    def is_superuser(self):
        return self.is_imis_admin



    @property
    def rights(self):
        rights = cache.get('rights_' + str(self.id))
        if rights:
            return rights
        rights = [rr.right_id for rr in RoleRight.filter_queryset().filter(
            role_id__in=[r.role_id for r in UserRole.filter_queryset().filter(
                user_id=self.id)]).distinct()]
        cache.set('rights_' + str(self.id), rights, timeout=None)
        return rights

    @property
    def rights_str(self):
        rights = [str(r) for r in self.rights]
        return rights

    @cached_property
    def health_facility(self):
        if self.health_facility_id:
            hf_model = apps.get_model("location", "HealthFacility")
            if hf_model:
                return hf_model.objects.filter(pk=self.health_facility_id).first()
        return None

    @property
    def is_officer(self):
        cache_name = f"user_eo_{self.login_name}"
        is_officer = cache.get(cache_name)
        if is_officer is None:
            is_officer = Officer.objects.filter(
                code=self.login_name,
                has_login=True,
                *filter_validity()
            ).exists()
            cache.set(cache_name, is_officer, None)
        return is_officer

    @property
    def is_claim_admin(self):
        # Unlike Officer ClaimAdmin model was moved to the claim module,
        # and it's not granted that the module is installed.
        if 'claim' in sys.modules:
            cache_name = f"user_ca_{self.login_name}"
            is_claim_admin = cache.get(cache_name)
            if is_claim_admin is None:
                
                from core.models.user import ClaimAdmin
                is_claim_admin = ClaimAdmin.objects.filter(
                    code=self.login_name,
                    has_login=True,
                    *filter_validity()
                ).exists()
                cache.set(cache_name, is_claim_admin, None)
            return is_claim_admin
        else:
            return False

    @property
    def is_imis_admin(self):
        is_admin = cache.get('is_admin_' + str(self.id))
        if is_admin is None:
            is_admin = Role.objects.filter(
                is_system=64,
                user_roles__user=self,
                validity_to__isnull=True,
                user_roles__validity_to__isnull=True,
                user_roles__user__validity_to__isnull=True
            ).exists()
            cache.set('is_admin_' + str(self.id), is_admin, 600)
        return is_admin

    def set_password(self, raw_password):
        from hashlib import sha256
        from secrets import token_hex
        validate_password(raw_password)
        self.private_key = token_hex(128)
        pwd_hash = sha256()
        pwd_hash.update(f"{raw_password.rstrip()}{self.private_key}".encode())
        self.password = (
            pwd_hash.hexdigest().upper()
        )  # Legacy requires this to be uppercase

    def check_password(self, raw_password):
        from hashlib import sha256

        pwd_hash = sha256()
        pwd_hash.update(f"{raw_password.rstrip()}{self.private_key}".encode())
        pwd_hash = pwd_hash.hexdigest()
        # logger.debug("pwd_hash %s -> %s, stored: %s", f"{raw_password.rstrip()}{self.private_key}", pwd_hash, self.password)
        # hashlib gives a lowercase digest while the legacy gives an uppercase one
        return pwd_hash == self.password.lower()

    @classmethod
    def is_interactive_user(cls, user):
        if isinstance(user, InteractiveUser):
            return user
        elif isinstance(user, User) and user.i_user is not None:
            return user.i_user
        else:
            return None

    @classmethod
    def get_email_field_name(cls):
        return "email"

    @classmethod
    def get_queryset(cls, queryset, user):
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        return queryset

    class Meta:
        managed = True
        db_table = 'tblUsers'


class UserRole(VersionedModel):
    id = models.AutoField(db_column='UserRoleID', primary_key=True)
    user = models.ForeignKey(
        InteractiveUser, models.DO_NOTHING, db_column='UserID', related_name="user_roles")
    role = models.ForeignKey(Role, models.DO_NOTHING,
                             db_column='RoleID', related_name="user_roles")
    audit_user_id = models.IntegerField(
        db_column='AudituserID', blank=True, null=True)

    class Meta:
        managed = True
        db_table = 'tblUserRole'


class User(UUIDModel, PermissionsMixin, UUIDVersionedModel):
    username = models.CharField(unique=True, max_length=50)
    t_user = models.ForeignKey(TechnicalUser, on_delete=models.CASCADE, blank=True, null=True)
    i_user = models.ForeignKey(InteractiveUser, on_delete=models.CASCADE, blank=True, null=True)
    officer = models.ForeignKey("Officer", on_delete=models.CASCADE, blank=True, null=True)
    claim_admin = models.ForeignKey("ClaimAdmin", on_delete=models.CASCADE, blank=True, null=True)

    USERNAME_FIELD = 'username'
    REQUIRED_FIELDS = []

    objects = UserManager()

    def save_history(self, **kwargs):
        # Prevent from saving history. It would lead to error due to username uniqueness.
        pass

    def delete_history(self, **kwargs):
        now = py_datetime.now()
        self.validity_from = now
        self.validity_to = now
        self.save()

    @property
    def _u(self):
        return self.i_user or self.officer or self.claim_admin or self.t_user

    def has_perms(self, perm_list, obj=None, list_evaluation_or=True):
        if not perm_list:
            return True
        if self.is_imis_admin:
            return True
        elif list_evaluation_or:
            return any(self.has_perm(perm, obj) for perm in perm_list)
        else:
            return super().has_perms(perm_list, obj)

    @property
    def id_for_audit(self):
        return self._u.id

    @property
    def last_login(self):
        return getattr(self._u, "last_login")

    @last_login.setter
    def last_login(self, value):
        return setattr(self._u, "last_login", value)

    @property
    def is_anonymous(self):
        return False

    @property
    def is_authenticated(self):
        return True

    @property
    def is_staff(self):
        return self._u.is_staff

    @property
    def is_superuser(self):
        return self._u.is_superuser

    @property
    def is_imis_admin(self):
        # 64 is system number for IMIS Administrator
        user = self._u
        if isinstance(user, InteractiveUser):
            return user.is_imis_admin
        else:
            return False

    @property
    def is_active(self):
        if self._u.validity_from is None and self._u.validity_to is None:
            return True
        now = py_datetime.now()
        if not self._u.validity_from is None and self._u.validity_from > now:
            return False
        if not self._u.validity_to is None and self._u.validity_to < now:
            return False
        return True

    def has_perm(self, perm, obj=None):
        i_user = self.i_user if obj is None else obj.i_user
        if i_user is not None and (
            i_user.is_superuser or 
            any(str(right) == perm for right in i_user.rights)
        ):
            return True
        else:
            return super(User, self).has_perm(perm, obj)

    @property
    def rights(self):
        if self.i_user:
            return self.i_user.rights
        return []

    def set_password(self, raw_password):
        if self._u and hasattr(self._u, "set_password"):
            return self._u.set_password(raw_password)
        self.clear_refresh_tokens()
        return None

    def clear_refresh_tokens(self):
        for refresh in self.refresh_tokens.filter(revoked__isnull=True):
            refresh.revoke()

    def get_session_auth_hash(self):
        key_salt = "core.User.get_session_auth_hash"
        return salted_hmac(key_salt, self.username).hexdigest()

    def get_health_facility(self):
        if self.claim_admin:
            return self.claim_admin.health_facility
        if self.i_user:
            return self.i_user.health_facility
        return None

    @property
    def health_facility(self):
        return self.get_health_facility()

    def __getattr__(self, name):
        if name == '_u':
            raise ValueError('wrapper has not been initialised')
        elif name == '__name__':
            return self.username
        elif name == 'get_session_auth_hash':
            return False
        elif hasattr(self._u, name):
            return getattr(self._u, name)
        elif name in self.__dict__:
            return self.__dict__[name]
        else:
            raise AttributeError(f"User has no attribute {name}")
        
    def __call__(self, *args, **kwargs):
        # if not self._u:
        #     raise ValueError('wrapper has not been initialised')
        if len(args) == 0 and len(kwargs) == 0 and not callable(self._u):
            # This happens when doing callable(user). Since this is a method, the class looks callable but it is not
            # To avoid this, we'll just return the object when calling it. This avoid issues in Django templates
            return self
        return self._u(*args, **kwargs)

    def __str__(self):
        if self.i_user:
            utype = 'i'
        elif self.t_user:
            utype = 't'
        elif self.officer:
            utype = 'o'
        elif self.claim_admin:
            utype = 'c'
        else:
            utype = '?'
        return "(%s) %s [%s]" % (utype, self.username, self.id)

    def save(self, *args, **kwargs):
        if self._u and self._u.id:
            self._u.save()
        super().save(*args, **kwargs)

    def shallow_save(self, *args, **kwargs):
        """ Unlike save(), shallow_save() won't attempt to save the subobjects, useful to avoid infinite recursion """
        super().save(*args, **kwargs)

    @classmethod
    def get_queryset(cls, queryset, user):
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            pass
        return queryset

    class Meta:
        managed = True
        db_table = 'core_User'


class UserGroup(models.Model):
    user = models.ForeignKey(User, models.DO_NOTHING)
    group = models.ForeignKey(Group, models.DO_NOTHING)

    class Meta:
        managed = False
        db_table = 'core_User_groups'
        unique_together = (('user', 'group'),)


class Officer(VersionedModel, ExtendableModel):
    id = models.AutoField(db_column='OfficerID', primary_key=True)
    uuid = models.CharField(db_column='OfficerUUID',
                            max_length=36, default=uuid.uuid4, unique=True)
    code = models.CharField(db_column='Code', max_length=50)
    last_name = models.CharField(db_column='LastName', max_length=100)
    other_names = models.CharField(db_column='OtherNames', max_length=100)
    dob = models.DateField(db_column='DOB', blank=True, null=True)
    phone = models.CharField(db_column='Phone', max_length=50, blank=True, null=True)
    location = models.ForeignKey('location.Location', models.DO_NOTHING, db_column='LocationId', blank=True, null=True)
    substitution_officer = models.ForeignKey('self', models.DO_NOTHING, db_column='OfficerIDSubst', blank=True,
                                             null=True)
    works_to = models.DateTimeField(db_column='WorksTo', blank=True, null=True)
    veo_code = models.CharField(db_column='VEOCode', max_length=50, blank=True, null=True)
    veo_last_name = models.CharField(db_column='VEOLastName', max_length=100, blank=True, null=True)
    veo_other_names = models.CharField(db_column='VEOOtherNames', max_length=100, blank=True, null=True)
    veo_dob = models.DateField(db_column='VEODOB', blank=True, null=True)
    veo_phone = models.CharField(db_column='VEOPhone', max_length=25, blank=True, null=True)
    audit_user_id = models.IntegerField(db_column='AuditUserID')
    # rowid = models.TextField(db_column='RowID', blank=True, null=True)   This field type is a guess.
    email = models.CharField(db_column='EmailId', max_length=200, blank=True, null=True)
    phone_communication = models.BooleanField(db_column='PhoneCommunication', blank=True, null=True)
    address = models.CharField(db_column="permanentaddress", max_length=100, blank=True, null=True)
    has_login = models.BooleanField(db_column='HasLogin', blank=True, null=True)

    # user = models.ForeignKey(User, db_column='UserID', blank=True, null=True, on_delete=models.CASCADE)

    def name(self):
        return " ".join(n for n in [self.last_name, self.other_names] if n is not None)

    def __str__(self):
        return "[%s] %s" % (self.code, self.name())

    @property
    def id_for_audit(self):
        return id

    @property
    def username(self):
        return self.code

    def get_username(self):
        return self.code

    @property
    def is_staff(self):
        return False

    @property
    def is_superuser(self):
        return False

    @cached_property
    def rights(self):
        return []

    @cached_property
    def rights_str(self):
        return []

    def set_password(self, raw_password):
        raise NotImplementedError("Shouldn't set a password on an Officer")

    def check_password(self, raw_password):
        return False

    @property
    def officer_allowed_locations(self):
        """
        Returns uuid of all locations allowed for given officer
        """
        from location.models import OfficerVillage, Location
        villages = OfficerVillage.objects\
            .filter(officer=self, validity_to__isnull=True)
        all_allowed_uuids = []
        for village in villages:
            allowed_uuids = [village.location.uuid]
            parent = village.location.parent
            while parent is not None:
                allowed_uuids.append(parent.uuid)
                parent = parent.parent
            all_allowed_uuids.extend(allowed_uuids)
        return Location.objects.filter(uuid__in=all_allowed_uuids)

    @classmethod
    def get_queryset(cls, queryset, user):
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        return queryset

    class Meta:
        managed = True
        db_table = 'tblOfficer'

    

def _get_default_expire_date():
    return py_datetime.now() + timedelta(days=1)


def _query_export_path(instance, filename):
    # file will be uploaded to MEDIA_ROOT/user_<id>/<filename>
    return F'query_exports/user_{instance.user.uuid}/{filename}'


class ClaimAdmin(VersionedModel):
    id = models.AutoField(db_column='ClaimAdminId', primary_key=True)
    uuid = models.CharField(db_column='ClaimAdminUUID',
                            max_length=36, default=uuid.uuid4, unique=True)

    code = models.CharField(db_column='ClaimAdminCode', max_length=50,
                            blank=True, null=True)
    last_name = models.CharField(
        db_column='LastName', max_length=100, blank=True, null=True)
    other_names = models.CharField(
        db_column='OtherNames', max_length=100, blank=True, null=True)
    dob = models.DateField(db_column='DOB', blank=True, null=True)
    email_id = models.CharField(
        db_column='EmailId', max_length=200, blank=True, null=True)
    phone = models.CharField(
        db_column='Phone', max_length=50, blank=True, null=True)
    health_facility = models.ForeignKey(
        "location.HealthFacility", models.DO_NOTHING, db_column='HFId', blank=True, null=True)
    has_login = models.BooleanField(
        db_column='HasLogin', blank=True, null=True)

    audit_user_id = models.IntegerField(
        db_column='AuditUserId', blank=True, null=True)
    # row_id = models.BinaryField(db_column='RowId', blank=True, null=True)

    def __str__(self):
        return self.code + " " + self.last_name + " " + self.other_names

    @classmethod
    def get_queryset(cls, queryset, user):
        queryset = cls.filter_queryset(queryset)
        # GraphQL calls with an info object while Rest calls with the user itself
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            from location.schema import LocationManager
            queryset = LocationManager().build_user_location_filter_query(
                user._u, prefix='health_facility__location', queryset=queryset, loc_types=['D'])
        return queryset

    @property
    def id_for_audit(self):
        return self.audit_user_id

    @property
    def username(self):
        return self.code

    def get_username(self):
        return self.code

    @property
    def is_staff(self):
        return False

    @property
    def is_superuser(self):
        return False

    def set_password(self, raw_password):
        raise NotImplementedError("Shouldn't set a password on an Officer")

    def check_password(self, raw_password):
        return False

    @property
    def officer_allowed_locations(self):
        """
        Returns uuid of all locations allowed for given officerLocationManager
        """
        district = self.health_facility.location
        all_allowed_uuids = [district.parent.uuid, district.uuid]
        child_locations = location_models.Location.objects.filter(
            parent=district).values_list('uuid', flat=True)
        while child_locations:
            all_allowed_uuids.extend(child_locations)
            child_locations = location_models.Location.objects\
                .filter(parent__uuid__in=child_locations)\
                .values_list('uuid', flat=True)

        return location_models.Location.objects.filter(uuid__in=all_allowed_uuids)

    class Meta:
        managed = True
        db_table = 'tblClaimAdmin'




from core.data_masking.abstract_mask import DataMaskAbs
from core.data_masking.masking_class_storage import MaskingClassStorage
from core.data_masking.masking_class_registry_point import MaskingClassRegistryPoint
from core.data_masking.masking_decorator import anonymize_gql



from django import forms
from django.contrib import admin
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.contrib.auth import password_validation, get_user_model
from django.contrib.auth.models import Group
from .models import TechnicalUser

User = get_user_model()


class TechnicalUserForm(forms.ModelForm):
    password = forms.CharField(
        strip=False,
        widget=forms.PasswordInput,
        help_text=password_validation.password_validators_help_text_html(),
    )

    class Meta:
        model = TechnicalUser
        fields = ['username', 'email', 'is_superuser', 'validity_from', 'validity_to']

    def save(self, commit=True):
        user = super().save(commit=False)
        user.set_password(self.cleaned_data["password"])
        user.is_staff = self.cleaned_data["is_superuser"]
        if commit:
            user.save()
        return user


class TechnicalUserAdmin(admin.ModelAdmin):
    form = TechnicalUserForm


class GroupAdminForm(forms.ModelForm):
    class Meta:
        model = Group
        exclude = []

    users = forms.ModelMultipleChoiceField(
        queryset=User.objects.all(),
        required=False,
        widget=FilteredSelectMultiple('users', False)
    )

    def __init__(self, *args, **kwargs):
        super(GroupAdminForm, self).__init__(*args, **kwargs)
        if self.instance.pk:
            self.fields['users'].initial = self.instance.user_set.all()

    def save_m2m(self):
        self.instance.user_set.set(self.cleaned_data['users'])

    def save(self, *args, **kwargs):
        instance = super(GroupAdminForm, self).save(commit=True)
        self.save_m2m()
        return instance


class GroupAdmin(admin.ModelAdmin):
    form = GroupAdminForm
    filter_horizontal = ['permissions']



# Generated by Django 3.2.18 on 2023-05-15 13:03

import core.fields
import datetime
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
        ('location', '0001_initial'),
        ('core', '0019_extended_field'),
    ]

    operations = [
        migrations.RenameField(
            model_name='role',
            old_name='altlanguage',
            new_name='alt_language',
        ),
        migrations.AddField(
            model_name='interactiveuser',
            name='is_associated',
            field=models.BooleanField(blank=True, db_column='IsAssociated', help_text='has a claim admin or enrolment officer account', null=True),
        ),
        migrations.AddField(
            model_name='interactiveuser',
            name='language',
            field=models.ForeignKey(db_column='LanguageID', on_delete=django.db.models.deletion.DO_NOTHING, to='core.language'),
        ),
        migrations.AddField(
            model_name='interactiveuser',
            name='last_login',
            field=models.DateTimeField(db_column='LastLogin', null=True),
        ),
        migrations.AddField(
            model_name='interactiveuser',
            name='password',
            field=models.CharField(blank=True, db_column='StoredPassword', help_text='By default a SHA256 of the private key (salt) and password', max_length=256, null=True),
        ),
        migrations.AddField(
            model_name='interactiveuser',
            name='password_validity',
            field=models.DateTimeField(blank=True, db_column='PasswordValidity', null=True),
        ),
        migrations.AddField(
            model_name='interactiveuser',
            name='private_key',
            field=models.CharField(blank=True, db_column='PrivateKey', help_text='The private key is actually a password salt', max_length=256, null=True),
        ),
        migrations.AddField(
            model_name='interactiveuser',
            name='uuid',
            field=models.CharField(db_column='UserUUID', default=uuid.uuid4, max_length=36, unique=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='address',
            field=models.CharField(blank=True, db_column='permanentaddress', max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='audit_user_id',
            field=models.IntegerField(db_column='AuditUserID'),
        ),
        migrations.AddField(
            model_name='officer',
            name='dob',
            field=models.DateField(blank=True, db_column='DOB', null=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='email',
            field=models.CharField(blank=True, db_column='EmailId', max_length=200, null=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='has_login',
            field=models.BooleanField(blank=True, db_column='HasLogin', null=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='json_ext',
            field=models.JSONField(blank=True, db_column='JsonExt', null=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='legacy_id',
            field=models.IntegerField(blank=True, db_column='LegacyID', null=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='location',
            field=models.ForeignKey(blank=True, db_column='LocationId', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='location.Location'),
        ),
        migrations.AddField(
            model_name='officer',
            name='phone',
            field=models.CharField(blank=True, db_column='Phone', max_length=50, null=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='phone_communication',
            field=models.BooleanField(blank=True, db_column='PhoneCommunication', null=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='substitution_officer',
            field=models.ForeignKey(blank=True, db_column='OfficerIDSubst', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='core.officer'),
        ),
        migrations.AddField(
            model_name='officer',
            name='validity_from',
            field=core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now),
        ),
        migrations.AddField(
            model_name='officer',
            name='validity_to',
            field=core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='veo_dob',
            field=models.DateField(blank=True, db_column='VEODOB', null=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='veo_last_name',
            field=models.CharField(blank=True, db_column='VEOLastName', max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='veo_other_names',
            field=models.CharField(blank=True, db_column='VEOOtherNames', max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='veo_phone',
            field=models.CharField(blank=True, db_column='VEOPhone', max_length=25, null=True),
        ),
        migrations.AddField(
            model_name='officer',
            name='works_to',
            field=models.DateTimeField(blank=True, db_column='WorksTo', null=True),
        ),
        migrations.AddField(
            model_name='usergroup',
            name='group',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='auth.group'),
        ),
        migrations.AddField(
            model_name='usergroup',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='core.user'),
        ),
        migrations.AddField(
            model_name='userrole',
            name='role',
            field=models.ForeignKey(db_column='RoleID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='user_roles', to='core.role'),
        ),
        migrations.AddField(
            model_name='userrole',
            name='user',
            field=models.ForeignKey(db_column='UserID', on_delete=django.db.models.deletion.DO_NOTHING, related_name='user_roles', to='core.interactiveuser'),
        ),
        migrations.AlterField(
            model_name='interactiveuser',
            name='login_name',
            field=models.CharField(db_column='LoginName', max_length=50),
        ),
        migrations.AlterField(
            model_name='interactiveuser',
            name='validity_from',
            field=core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now),
        ),
        migrations.AlterField(
            model_name='language',
            name='code',
            field=models.CharField(db_column='LanguageCode', max_length=5, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='officer',
            name='code',
            field=models.CharField(db_column='Code', max_length=50),
        ),
        migrations.AlterField(
            model_name='officer',
            name='veo_code',
            field=models.CharField(blank=True, db_column='VEOCode', max_length=50, null=True),
        ),
        migrations.AlterField(
            model_name='role',
            name='uuid',
            field=models.CharField(db_column='RoleUUID', default=uuid.uuid4, max_length=36, unique=True),
        ),
        migrations.AlterField(
            model_name='role',
            name='validity_from',
            field=core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now),
        ),
        migrations.AlterField(
            model_name='role',
            name='validity_to',
            field=core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True),
        ),
        migrations.AlterField(
            model_name='roleright',
            name='validity_from',
            field=core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now),
        ),
        migrations.AlterField(
            model_name='roleright',
            name='validity_to',
            field=core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True),
        ),
        migrations.AlterField(
            model_name='userrole',
            name='validity_from',
            field=core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now),
        ),
        migrations.AlterField(
            model_name='userrole',
            name='validity_to',
            field=core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True),
        ),
        migrations.AlterUniqueTogether(
            name='usergroup',
            unique_together={('user', 'group')},
        ),
        migrations.AlterModelTable(
            name='usergroup',
            table='core_User_groups',
        ),
    ]



# Generated by Django 2.1.11 on 2019-10-25 12:00

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0007_auto_20191008_0923'),
    ]

    operations = [
        migrations.CreateModel(
            name='Officer',
            fields=[
                ('id', models.AutoField(db_column='OfficerID', primary_key=True, serialize=False)),
                ('uuid', models.CharField(db_column='OfficerUUID', default=uuid.uuid4, max_length=36, unique=True)),
                ('code', models.CharField(db_column='Code', max_length=8)),
                ('last_name', models.CharField(db_column='LastName', max_length=100)),
                ('other_names', models.CharField(db_column='OtherNames', max_length=100)),
            ],
            options={
                'db_table': 'tblOfficer',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='Role',
            fields=[
                ('id', models.AutoField(db_column='RoleID', primary_key=True, serialize=False)),
                ('uuid', models.CharField(db_column='RoleUUID', max_length=36)),
                ('name', models.CharField(db_column='RoleName', max_length=50)),
                ('altlanguage', models.CharField(blank=True, db_column='AltLanguage', max_length=50, null=True)),
                ('is_system', models.IntegerField(db_column='IsSystem')),
                ('is_blocked', models.BooleanField(db_column='IsBlocked')),
                ('validity_from', models.DateTimeField(db_column='ValidityFrom')),
                ('validity_to', models.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('audit_user_id', models.IntegerField(blank=True, db_column='AuditUserID', null=True)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
            ],
            options={
                'db_table': 'tblRole',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='RoleRight',
            fields=[
                ('id', models.AutoField(db_column='RoleRightID', primary_key=True, serialize=False)),
                ('right_id', models.IntegerField(db_column='RightID')),
                ('validity_from', models.DateTimeField(db_column='ValidityFrom')),
                ('validity_to', models.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('audit_user_id', models.IntegerField(blank=True, db_column='AuditUserId', null=True)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
            ],
            options={
                'db_table': 'tblRoleRight',
                'managed': False,
            },
        ),
        migrations.CreateModel(
            name='UserRole',
            fields=[
                ('id', models.AutoField(db_column='UserRoleID', primary_key=True, serialize=False)),
                ('validity_from', models.DateTimeField(db_column='ValidityFrom')),
                ('validity_to', models.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('audit_user_id', models.IntegerField(blank=True, db_column='AudituserID', null=True)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
            ],
            options={
                'db_table': 'tblUserRole',
                'managed': False,
            },
        ),
    ]



import ast
import json
import logging
import uuid
from importlib import import_module
from typing import Any, Dict, Type

import core
import graphene
import jsonschema
from django.db import models
from django.apps import AppConfig
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import PermissionDenied, ValidationError
from django.core.files.storage import default_storage
from django.db.models import Q
from django.http import FileResponse
from django.utils.translation import gettext as _
from graphql import GraphQLError
from password_validator import PasswordValidator
from zxcvbn import zxcvbn
import datetime
from django.core.cache import caches

logger = logging.getLogger(__file__)

UNIQUE_FIELDS = ('pk', 'id', 'uuid')
CACHE_TIMEOUT = 3600 * 24
cache = caches["default"]

__all__ = [
    "TimeUtils",
    "full_class_name",
    "comparable",
    "filter_validity",
    "prefix_filterset",
    "assert_string_length",
    "PATIENT_CATEGORY_MASK_MALE",
    "PATIENT_CATEGORY_MASK_FEMALE",
    "PATIENT_CATEGORY_MASK_ADULT",
    "PATIENT_CATEGORY_MASK_MINOR",
    "patient_category_mask",
    "ExtendedConnection",
    "get_scheduler_method_ref",
    "ExtendedRelayConnection",
]


class TimeUtils(object):

    @classmethod
    def now(cls):
        return core.datetime.datetime.now()

    @classmethod
    def date(cls):
        return core.datetime.datetime.date(cls.now())


def full_class_name(o):
    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__
    return module + "." + o.__class__.__name__


def comparable(cls):
    """Class decorator providing generic comparison functionality"""

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    cls.__eq__ = __eq__
    cls.__ne__ = __ne__
    return cls


def filter_validity(arg="validity", prefix="", **kwargs):
    
    validity = kwargs.get(arg)
    if validity is None:
        return [Q(**{f"{prefix}validity_to__isnull": True})]
    elif isinstance(validity, str):
        validity = datetime.datetime.strptime(validity)
    validity = datetime.datetime(validity.year, validity.month, validity.day, 23, 59, 59)
    return [
        Q(**{f"{prefix}validity_from__lte": validity}),
        Q(**{f"{prefix}validity_to__isnull": True}) | Q(**{f"{prefix}validity_to__gte": validity}),
    ]


def filter_validity_business_model(
        arg="dateValidFrom__Gte", arg2="dateValidTo__Lte", **kwargs
):
    date_valid_from = kwargs.get(arg)
    date_valid_to = kwargs.get(arg2)
    # default scenario
    if not date_valid_from and not date_valid_to:
        today = core.datetime.datetime.now()
        return __place_the_filters(date_start=today, date_end=None)

    # scenario - only date valid to set
    if not date_valid_from and date_valid_to:
        today = core.datetime.datetime.now()
        oldest = min([today, date_valid_to])
        return __place_the_filters(date_start=oldest, date_end=date_valid_to)

    # scenario - only date valid from
    if date_valid_from and not date_valid_to:
        return __place_the_filters(date_start=date_valid_from, date_end=None)

    # scenario - both filters set
    if date_valid_from and date_valid_to:
        return __place_the_filters(date_start=date_valid_from, date_end=date_valid_to)


def __place_the_filters(date_start, date_end):
    """funtion related to 'filter_validity_business_model'
    function so as to set up the chosen filters
    to filter the validity of the entity
    """
    if not date_end:
        return (
            Q(date_valid_from__isnull=False),
            Q(date_valid_to__isnull=True) | Q(date_valid_to__gte=date_start),
        )
    return (
        Q(date_valid_from__lte=date_end),
        Q(date_valid_to__isnull=True) | Q(date_valid_to__gte=date_start),
    )


def append_validity_filter(**kwargs):
    default_filter = kwargs.get("applyDefaultValidityFilter", False)
    date_valid_from = kwargs.get("dateValidFrom__Gte", None)
    date_valid_to = kwargs.get("dateValidTo__Lte", None)
    filters = []
    # check if we can use default filter validity
    if date_valid_from is None and date_valid_to is None:
        if default_filter:
            filters = [*filter_validity_business_model(**kwargs)]
        else:
            filters = []
    else:
        filters = [*filter_validity_business_model(**kwargs)]
    return filters


def flatten_dict(d, parent_key="", sep="_"):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def filter_is_deleted(arg="is_deleted", **kwargs):
    is_deleted = kwargs.get(arg)
    if is_deleted is None:
        is_deleted = False
    return Q(is_deleted=is_deleted)


def prefix_filterset(prefix, filterset):
    if type(filterset) is dict:
        return {(prefix + k): v for k, v in filterset.items()}
    elif type(filterset) is list:
        return [(prefix + x) for x in filterset]
    else:
        return filterset


def assert_string_length(str_value, length):
    if length and len(str_value) > length:
        raise Exception(
            _("core.string.over_max_length")
            % {"value": str_value, "max_length": length}
        )


PATIENT_CATEGORY_MASK_MALE = 1
PATIENT_CATEGORY_MASK_FEMALE = 2
PATIENT_CATEGORY_MASK_ADULT = 4
PATIENT_CATEGORY_MASK_MINOR = 8


def patient_category_mask(insuree, target_date):
    if type(target_date) is str:
        from core import datetime

        # TODO: this should be nicer
        target_date = datetime.date(*[int(x) for x in target_date.split("-")])
    mask = 0
    if not insuree.gender:
        raise NotImplementedError(_("core.insuree.unknown_gender"))
    if not insuree.dob:
        raise NotImplementedError(_("core.insuree.unknown_dob"))

    if insuree.gender.code in ("M", "O"):
        mask = mask | PATIENT_CATEGORY_MASK_MALE
    else:
        mask = mask | PATIENT_CATEGORY_MASK_FEMALE

    if insuree.is_adult(target_date):
        mask = mask | PATIENT_CATEGORY_MASK_ADULT
    else:
        mask = mask | PATIENT_CATEGORY_MASK_MINOR
    return mask


class CachedManager(models.Manager):
    
    def get(self, *args, **kwargs):
        """
        Overrides the get() method to check Redis cache before
        performing a DB lookup for simple unique lookups.
        """
        
        cache_key = None
        value = None
        value_key = None
        # Case 1: Simple kwargs lookup.
        if kwargs and len(kwargs) == 1:
            key = list(kwargs.keys())[0]
            if key in UNIQUE_FIELDS:
                value = kwargs[key]
                value_key = key
        # use case for Family Request elements in args
        elif not kwargs and args and len(args) == 1 :
            if len(args[0].children) == 1:
                field, arg_value = args[0].children[0]
                if field in UNIQUE_FIELDS:
                    value = arg_value
                    value_key = 'pk'
        # checking if the key is the PK

        if value_key == 'pk':
            cache_key = get_cache_key(self.model, value)
        elif value:
            field = self.model._meta.get_field(value_key) 
            if not field.primary_key:
                # we get the unique fields / pk conversion if exist
                value = get_cache_key(self.model, kwargs[key])
        # Convert UUID objects to string for the cache key.
            if isinstance(value, uuid.UUID):
                value = str(value)
            else:
                try:
                    # Convert to int if possible.
                    value = int(value)
                except (ValueError, TypeError):
                    pass
            cache_key = get_cache_key(self.model, value)
        # If we constructed a cache key, try to retrieve from the cache.
        if cache_key:
            cached_instance = cache.get(cache_key)
            if cached_instance is not None:
                if not isinstance(cached_instance, self.model):
                    logger.error("wrong model for cached instance for key: %s", cache_key)
                logger.debug("Returning cached instance for key: %s", cache_key)
                return cached_instance

            # Not in cache; perform DB lookup.
        instance = super().get(*args, **kwargs)
        cache.set(cache_key, instance, timeout=None)
        logger.debug("Cached instance %s after DB lookup", cache_key)
        return instance




class CachedModelMixin():
    def update_cache(self):
        """
        Updates the cache for this object after saving.
        """
        cache.set(get_cache_key(self.__class__, self.pk), self,  timeout=CACHE_TIMEOUT)
        for f in UNIQUE_FIELDS:

            # get_field raised an error on property raise 
            if self.pk != getattr(self,f, self.pk):
                cache.set(get_cache_key(self.__class__, getattr(self,f)), self.pk,  timeout=CACHE_TIMEOUT)

        logger.debug("Saved and cached instance: %s", self)
        
    def delete_cache(self):
        """
        Deletes the cache entry for this object.
        """
        cache_key = f"{self.__class__.__name__}:{self.pk}"
        cache.delete(cache_key)
        logger.debug(f"Removed instance from cache: {cache_key}")

class ExtendedConnection(graphene.Connection):
    """
    Adds total_count and edge_count to Graphene connections. To use, simply add to the
    Graphene object definition Meta:
    `connection_class = ExtendedConnection`
    """

    class Meta:
        abstract = True

    total_count = graphene.Int()
    edge_count = graphene.Int()

    def resolve_total_count(self, info, **kwargs):
        if not info.context.user.is_authenticated:
            raise PermissionDenied(_("unauthorized"))
        return self.length

    def resolve_edge_count(self, info, **kwargs):
        if not info.context.user.is_authenticated:
            raise PermissionDenied(_("unauthorized"))
        return len(self.edges)


def block_update(update_dict, current_object, attribute_name, Ex=ValueError):
    if attribute_name in update_dict and update_dict["code"] != getattr(
            current_object, attribute_name
    ):
        raise Ex("That {attribute_name} field is not editable")


def get_scheduler_method_ref(name):
    """
    Use to retrieve the method reference from a str name. This is necessary when the module cannot be imported from
    that location.
    :param name: claim.module.submodule.method or similar name
    :return: reference to the method
    """
    split_name = name.split(".")
    module = __import__(".".join(split_name[:-1]))
    for subitem in split_name[1:]:
        module = getattr(module, subitem)
    return module


class ExtendedRelayConnection(graphene.relay.Connection):
    """
    Adds total_count and edge_count to Graphene Relay connections.
    """

    class Meta:
        abstract = True

    total_count = graphene.Int()
    edge_count = graphene.Int()

    def resolve_total_count(self, info, **kwargs):
        return len(self.iterable)

    def resolve_edge_count(self, info, **kwargs):
        return len(self.edges)


def get_first_or_default_language():
    from core.models import Language

    sorted_languages = Language.objects.filter(sort_order__isnull=False)
    if sorted_languages.exists():
        return sorted_languages.order_by("sort_order").first()
    else:
        return Language.objects.first()


def insert_role_right_for_system(system_role, right_id, apps):
    RoleRight = apps.get_model("core", "RoleRight")
    Role = apps.get_model("core", "Role")
    existing_roles = Role.objects.filter(
        is_system=system_role, validity_to__isnull=True
    )
    if not existing_roles:
        logger.warning(
            "Migration requested a role_right for system role %s but couldn't find that role",
            system_role,
        )
    else:

        for existing_role in existing_roles:
            role_rights = RoleRight.objects.filter(
                role=existing_role, right_id=right_id
            ).first()
            if not role_rights:
                RoleRight.objects.create(
                    role=existing_role,
                    right_id=right_id,
                    validity_from=datetime.datetime.now()
                )



def remove_role_right_for_system(system_role, right_id, apps):
    RoleRight = apps.get_model("core", "RoleRight")
    Role = apps.get_model("core", "Role")
    existing_roles = Role.objects.filter(
        is_system=system_role, validity_to__isnull=True
    )
    if not existing_roles:
        logger.warning(
            "Migration requested to remove a role_right for system role %s but couldn't find that role",
            system_role,
        )
    for existing_role in existing_roles:
        role_rights = RoleRight.objects.filter(role=existing_role, right_id=right_id)
        if not role_rights:
            logger.warning(
                "Role right not found for system role %s and right ID %s",
                system_role,
                right_id,
            )
        for role_right in role_rights:
            role_right.delete()
            logger.info(
                "Role right removed for system role %s and right ID %s",
                system_role,
                right_id,
            )


def convert_to_python_value(string):
    try:
        value = ast.literal_eval(string)
        return value
    except (SyntaxError, ValueError):
        return string


def is_valid_uuid(string):
    try:
        uuid.UUID(str(string))
        return True
    except ValueError:
        return False


def validate_json_schema(schema):
    try:
        if not isinstance(schema, dict):
            schema = json.loads(schema)
        jsonschema.Draft7Validator.check_schema(schema)
        return []
    except jsonschema.exceptions.SchemaError as schema_error:
        return [
            {"message": _("core.utils.schema_validation.invalid_schema") % {"error": str(schema_error)}}
        ]
    except ValueError as json_error:
        return [
            {"message": _("core.utils.schema_validation.invalid_json") % {"error": str(json_error)}}
        ]


def to_json_safe_value(value):
    try:
        json.dumps(value)
        return value
    except (TypeError, ValueError):
        return str(value)


class CustomPasswordValidator:
    def __init__(self, uppercase=0, lowercase=0, digits=0, symbols=0):
        self.schema = PasswordValidator()
        self.requirements = {
            "PASSWORD_UPPERCASE": "uppercase",
            "PASSWORD_LOWERCASE": "lowercase",
            "PASSWORD_DIGITS": "digits",
            "PASSWORD_SYMBOLS": "symbols",
        }
        self.set_password_policy()

    def set_password_policy(self):
        self.schema.min(settings.PASSWORD_MIN_LENGTH)
        for setting, method in self.requirements.items():
            if getattr(settings, setting) > 0:
                getattr(self.schema.has(), method)()

    def validate(self, password, user=None):
        if not self.schema.validate(password):
            raise ValidationError(self.get_help_text())
        zxcvbn_result = zxcvbn(password)
        if zxcvbn_result["score"] < 3:
            raise ValidationError(
                "Password is too weak. Avoid common patterns and dictionary words."
            )

    def get_help_text(self):
        return (
            f"Your password must be at least {settings.PASSWORD_MIN_LENGTH} characters long, "
            f"contain at least {settings.PASSWORD_UPPERCASE} uppercase letter(s), "
            f"{settings.PASSWORD_LOWERCASE} lowercase letter(s), "
            f"{settings.PASSWORD_DIGITS} number(s), and "
            f"{settings.PASSWORD_SYMBOLS} special character(s)."
        )


class DefaultStorageFileHandler:
    def __init__(self, file_path):
        self.file_path = file_path

    def save_file(self, file):
        self.check_file_path()
        default_storage.save(self.file_path, file)
        file.seek(0)

    def remove_file(self):
        if default_storage.exists(self.file_path):
            default_storage.delete(self.file_path)

    def get_file_content(self):
        if not default_storage.exists(self.file_path):
            raise FileNotFoundError("File does not exist at the specified path.")
        with default_storage.open(self.file_path, "rb") as source:
            return source.read()

    def get_file_response_csv(self, file_name=None):
        if not default_storage.exists(self.file_path):
            raise FileNotFoundError("File does not exist at the specified path.")
        response = FileResponse(default_storage.open(self.file_path, "rb"))
        response["Content-Type"] = "text/csv"
        response["Content-Disposition"] = (
            f'attachment; filename="{file_name if file_name else "default.csv"}"'
        )
        return response

    def check_file_path(self):
        if default_storage.exists(self.file_path):
            raise FileExistsError("File already exists at the specified path.")

    @staticmethod
    def list_files(directory):
        """
        Get a list of files in the specified directory within default storage.
        """
        return default_storage.listdir(directory)


class ConfigUtilMixin:
    @classmethod
    def _load_config_fields(cls, default_cfg: Dict[str, Any]):
        """
        Load all config fields that match current AppConfig class fields, all custom fields have to be loaded separately
        """
        for field in default_cfg:
            if hasattr(cls, field):
                setattr(cls, field, default_cfg[field])

    @classmethod
    def _load_config_function(cls, function_name, path):
        """
        Load a function specified as module path into config.
        Example:
        "core.apps.function" will be loaded as "from core.apps import function" and assigned as "cls.function_name"
        """
        try:
            mod, name = path.rsplit(".", 1)
            if not mod or not name:
                raise ImportError(
                    "Invalid function path, module and function name are required"
                )
            module = import_module(mod)
            function = getattr(module, name)
            setattr(cls, function_name, function)
        except ImportError as e:
            logger.error(
                "Failed to configure function '%s' as '%s.%s': %s",
                path,
                cls.__name__,
                function_name,
                str(e),
            )


def clear_cache(instance):
    cache.delete(get_cache_key(instance.__class__, instance.pk))


def get_cache_key(model, id):
    return f"cs_{model.__name__}_{id}"


def is_this_session_superuser(session_key):
    from django.contrib.sessions.models import Session
    from django.utils.timezone import now
    from core.models import User

    try:
        session = Session.objects.get(session_key=session_key, expire_date__gte=now())
        data = session.get_decoded()
        user_id = data.get('_auth_user_id')
        if user_id:
            user = User.objects.get(id=user_id)
            if user.is_superuser:
                return True
    except Session.DoesNotExist:
        pass
    except Exception as e:
        pass

    return False



import logging
import sys
import uuid
from datetime import timedelta, datetime as py_datetime
from django.core.cache import cache
from cached_property import cached_property
from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import AbstractBaseUser, BaseUserManager, PermissionsMixin, Group
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.db import models
from django.utils.crypto import salted_hmac
from graphql import ResolveInfo
import core
#from core.datetimes.ad_datetime import datetime as py_datetime
from django.conf import settings

from ..utils import filter_validity
from .base import *
from .user import *
from .base_mutation import *
from .versioned_model import *

logger = logging.getLogger(__name__)



class RoleMutation(UUIDModel, ObjectMutation):
    role = models.ForeignKey(Role, models.DO_NOTHING,
                             related_name='mutations')
    mutation = models.ForeignKey(
        MutationLog, models.DO_NOTHING, related_name='roles')

    class Meta:
        managed = True
        db_table = "core_RoleMutation"


class UserMutation(UUIDModel, ObjectMutation):
    core_user = models.ForeignKey(User, models.CASCADE, related_name='mutations')
    mutation = models.ForeignKey(MutationLog, models.DO_NOTHING, related_name='users')

    class Meta:
        managed = True
        db_table = "core_UserMutation"




# Generated by Django 4.2.9 on 2024-02-07 16:47

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0026_mutationlog_autogenerated_code'),
    ]

    operations = [
        migrations.AlterField(
            model_name='interactiveuser',
            name='last_login',
            field=models.DateTimeField(blank=True, db_column='LastLogin', null=True),
        ),
        migrations.AlterField(
            model_name='technicaluser',
            name='validity_from',
            field=models.DateTimeField(blank=True, default=datetime.datetime.now, null=True),
        ),
    ]



import re
from io import StringIO

from django.core.management import call_command
from django.db import models, transaction, connection
from django.db.models import F
from django.db.models.functions import Replace
from simple_history.utils import get_history_model_for_model

from pathlib import Path
import os

from core.models import HistoryModel


class temporary_module_migration(object):
    def __init__(self, module_name):
        self.module_name = module_name
        self._check_if_module_is_local()

    def __enter__(self):
        self.migration_name = self.__create_migration()
        return self.migration_name

    def __exit__(self, exception_type, exception_value, traceback):
        """
        Delete migration file from relative path.
        """

        migrations_path = self._get_migrations_path()
        file = os.path.join(migrations_path, F'{self.migration_name}.py')
        os.remove(file)

        if exception_type:
            raise exception_type(exception_value)

    def __create_migration(self):
        out = StringIO()
        call_command('makemigrations', self.module_name, stdout=out)
        out = out.getvalue()
        x = re.compile("\/migrations\/(.*)\.py")
        migration_name = x.search(out).groups()[0]
        return migration_name

    def _check_if_module_is_local(self):
        migrations_path = self._get_migrations_path()
        if not os.path.isdir(migrations_path):
            raise ValueError(
                "Temporary migration can be created only for modules installed locally next to openimis-be_py."
                "Core module also has to be installed locally. "
                f"Either module {self.module_name} is not installed locally or it doesn't provide /migrations/ folder."
            )

    def _get_migrations_path(self):
        path_dir = Path(__file__).parent.parent.parent.parent
        relative_path = F'openimis-be-{self.module_name}_py/{self.module_name}/migrations/'
        return os.path.join(path_dir, relative_path)


class CreateTempTable:
    def __init__(self, module, model, type_of_history_model):
        self.module = module
        self.model = model
        self.type_of_history_model = type_of_history_model
        if self.type_of_history_model not in ['HISTORY_MODEL', 'BUSINESS_HISTORY_MODEL']:
            raise AttributeError("Invalid new model field type")

        versioned_model_fields = self._get_versioned_model_fields()
        self.HistoryTransitionModel = type('HistoryTransitionModel', (HistoryModel,), {
            **versioned_model_fields,
            '__module__': self.module,
        })

    @transaction.atomic
    def migrate_to_history_model(self):
        sql_queries = self._create_transition_queries()
        self.execute_sqls(sql_queries)
        print("--______________________________________________")
        print("--______________________________________________")
        print(rf"{sql_queries}")
        print("--______________________________________________")
        print("--______________________________________________")
        raise ValueError("Rollback")

    def _create_transition_queries(self):
        create_table_sql = self.migrate()
        current, historical = self.split_historical_data()
        insert_sql = self.load_current(current)
        insert_history_sql = self.load_historical(historical)
        return [
            create_table_sql,
            insert_sql,
            insert_history_sql
        ]

    def _get_versioned_model_fields(self):
        new_fields = {}
        # Add existing model fields
        data_fields = [x for x in self.model._meta.get_fields() if isinstance(x, models.fields.Field)]
        for f in data_fields:
            name = F"versioned_{f.name}"
            if name == 'versioned_id':
                # PK is replaced not transitioned
                continue

            f = f.clone()
            f.name = name
            f.db_column = name
            f.is_relation = False
            new_fields[name] = f
        return new_fields

    def migrate(self):
        with temporary_module_migration(self.module) as migration_name:
            out = StringIO()
            call_command('sqlmigrate', self.module, migration_name, stdout=out)
            out = out.getvalue()
            x = re.compile("(CREATE TABLE .*\;)")
            sql_create_table = x.findall(out)
            query = '\n;'.join(sql_create_table)
            return query, []

    def split_historical_data(self):
        historical = self.model.objects.filter(validity_to__isnull=False, legacy_id__isnull=False).all()
        current = self.model.objects.filter(legacy_id__isnull=True).all()
        return self.wrap_current(current), self.annotate_historical_fields(self.wrap_current(historical))

    def execute_sqls(self, sqls):
        with connection.cursor() as cursor:
            for sql, params in sqls:
                cursor.execute(sql, params)
            cursor.execute("SELECT COUNT(*) FROM insuree_historytransitionmodel")
            row = cursor.fetchone()
            return row

    def get_history_models_annotation(self):
        versioned_model_fields = self._get_versioned_model_fields()
        extra = {
            'UserUpdatedUUID': Replace('versioned_uuid', models.Value('-'), models.Value('')),  # For now fixed
            'UUID': Replace('versioned_uuid', models.Value('-'), models.Value('')),
            'UserCreatedUUID': Replace('versioned_uuid', models.Value('-'), models.Value('')),
            'DateCreated': F('versioned_validity_from'),
            'version': models.Value(0),
            'Json_ext': F('versioned_json_ext'),
            'DateUpdated': F('versioned_validity_from'),
            'isDeleted': models.Value(False)
        }
        return [*list(versioned_model_fields.keys()), *list(extra.keys())]

    def get_historical_models_annotation(self):
        history = self.get_history_models_annotation()
        extra = {
            'history_date': F('validity_to'),
            'history_change_reason': models.Value('Versioned model migration'),
            'history_type': models.Value('+'),
            'history_user_id': Replace('versioned_uuid', models.Value('-'), models.Value('')),
        }
        return [*history, *list(extra.keys())]

    def annotate_values(self, queryset):
        versioned_model_fields = self._get_versioned_model_fields()
        aliases = {x.replace('versioned_', ''): x for x in versioned_model_fields.keys()}
        return queryset \
            .annotate(**{transition_name: F(versioned_name) for versioned_name, transition_name in aliases.items()}) \
            .all()

    def annotate_history_model_fields(self, base):
        extra = {
            'UserUpdatedUUID': Replace('versioned_uuid', models.Value('-'), models.Value('')),  # For now fixed
            'UUID': Replace('versioned_uuid', models.Value('-'), models.Value('')),
            'UserCreatedUUID': Replace('versioned_uuid', models.Value('-'), models.Value('')),
            'DateCreated': F('versioned_validity_from'),
            'version': models.Value(0),
            'Json_ext': F('versioned_json_ext'),
            'DateUpdated': F('versioned_validity_from'),
            'isDeleted': models.Case(models.When(validity_to__isnull=True, then=models.Value(False)), default=True)
        }
        base = base.annotate(**extra)
        return base

    def annotate_historical_fields(self, base):
        extra = {
            'history_date': F('validity_to'),
            'history_change_reason': models.Value('Versioned model migration'),
            'history_type': models.Value('+'),
            'history_user_id': Replace('versioned_uuid', models.Value('-'), models.Value('')),
        }
        base = base.annotate(**extra)
        return base

    def wrap_current(self, queryset):
        base = self.annotate_values(queryset)
        base = self.annotate_history_model_fields(base)
        values = self.get_history_models_annotation()
        return base.values(*values)

    def load_current(self, data):
        new_query, params = data.query.sql_with_params()
        columns = self.get_history_models_annotation()
        return F'INSERT INTO {self.HistoryTransitionModel._meta.db_table}({", ".join(columns)}) {new_query}', params

    def load_historical(self, data):
        history_table = get_history_model_for_model(self.HistoryTransitionModel)
        new_query, params = data.query.sql_with_params()
        columns = self.get_historical_models_annotation()
        return F'INSERT INTO {history_table._meta.db_table}({", ".join(columns)}) {new_query}', params



import abc
import datetime


class AbsStrategy(object, metaclass=abc.ABCMeta):

    @classmethod
    def get_version(cls):
        return type(cls)._version

    @classmethod
    def set_version(cls, val):
        type(cls)._version = val

    version = property(get_version, set_version)

    @classmethod
    def get_uuid(cls):
        return type(cls)._uuid

    @classmethod
    def set_uuid(cls, val):
        type(cls)._uuid = val

    uuid = property(get_uuid, set_uuid)

    @classmethod
    def get_calculation_rule_name(cls):
        return type(cls)._calculation_rule_name

    @classmethod
    def set_calculation_rule_name(cls, val):
        type(cls)._calculation_rule_name = val

    calculation_rule_name = property(get_calculation_rule_name, set_calculation_rule_name)

    @classmethod
    def get_description(cls):
        return type(cls)._description

    @classmethod
    def set_description(cls, val):
        type(cls)._description = val

    description = property(get_description, set_description)

    @classmethod
    def get_impacted_class_parameter(cls):
        return type(cls)._impacted_class_parameter

    @classmethod
    def set_impacted_class_parameter(cls, val):
        type(cls)._impacted_class_parameter = val

    impacted_class_parameter = property(get_impacted_class_parameter, set_impacted_class_parameter)

    @classmethod
    def get_type(cls):
        return type(cls)._type

    @classmethod
    def set_type(cls, val):
        type(cls)._type = val

    type = property(get_type, set_type)

    @classmethod
    def get_sub_type(cls):
        return type(cls)._sub_type

    @classmethod
    def set_sub_type(cls, val):
        type(cls)._sub_type = val

    sub_type = property(get_sub_type, set_sub_type)

    @classmethod
    def get_from_to(cls):
        return type(cls)._from_to

    @classmethod
    def set_from_to(cls, val):
        type(cls)._from_to = val

    from_to = property(get_from_to, set_from_to)

    @classmethod
    def ready(cls):
        now = datetime.datetime.now()
        condition_is_valid = (now >= cls.date_valid_from and now <= cls.date_valid_to) \
            if cls.date_valid_to else (now >= cls.date_valid_from and cls.date_valid_to is None)
        if not condition_is_valid:
            cls.status = "inactive"

    @classmethod
    @abc.abstractmethod
    def check_calculation(cls, instance):
        return

    @classmethod
    @abc.abstractmethod
    def active_for_object(cls, instance, context, type, sub_type):
        return

    @classmethod
    @abc.abstractmethod
    def calculate(cls, instance, *args, **kwargs):
        return

    @classmethod
    def get_linked_class(cls, sender, class_name, **kwargs):
        # calculation are loaded on the side, therefore contentType have to be loaded on execution
        from django.contrib.contenttypes.models import ContentType
        list_class = []
        if class_name is not None:
            model_class = ContentType.objects.filter(model__iexact=class_name).first()
            if model_class:
                model_class = model_class.model_class()
                list_class = list_class + \
                             [f.remote_field.model.__name__ for f in model_class._meta.fields
                              if f.get_internal_type() == 'ForeignKey' and f.remote_field.model.__name__ != "User"]
        else:
            list_class.append("Calculation")
        return list_class

    @classmethod
    @abc.abstractmethod
    def convert(cls, instance, convert_to, **kwargs):
        return

    @classmethod
    def get_rule_name(cls, sender, class_name, **kwargs):
        for object_class in cls.impacted_class_parameter:
            if object_class["class"] == class_name:
                return cls

    @classmethod
    def get_rule_details(cls, sender, class_name, **kwargs):
        for object_class in cls.impacted_class_parameter:
            if object_class["class"] == class_name:
                return object_class

    @classmethod
    def get_parameters(cls, sender, class_name, instance, **kwargs):
        """
            class_name is the class name of the object where the calculation param need to be added
            instance is where the link with a calculation need to be found,
            like the CPB in case of PH insuree or Contract Details
            return a list only with rule details that matches step 1 and 2
        """
        rule_details = cls.get_rule_details(sender=sender, class_name=class_name)
        if rule_details:
            if cls.check_calculation(instance=instance):
                return rule_details["parameters"] if "parameters" in rule_details else []

    @classmethod
    def run_calculation_rules(cls, sender, instance, user, context, **kwargs):
        """
            this function will send a signal and the rules will
            reply if they have object matching the classname in their list of object
        """
        list_class = cls.get_linked_class(sender, instance.__class__.__name__)
        # if the class have a calculation param, (like contribution or payment plan) add class name
        if hasattr(instance, 'calculation'):
            list_class.append(instance.__class__.__name__)
        if list_class:
            for class_name in list_class:
                rule_details = cls.get_rule_details(class_name=class_name, sender=sender)
                if rule_details or len(cls.impacted_class_parameter) == 0:
                    # add context to kwargs
                    kwargs["context"] = context
                    result = cls.calculate_if_active_for_object(instance, **kwargs)
                    return result

    @classmethod
    def calculate_if_active_for_object(cls, instance, **kwargs):
        if cls.active_for_object(instance=instance, context=kwargs['context']):
            return cls.calculate(instance, **kwargs)

    @classmethod
    def run_convert(cls, instance, convert_to, **kwargs):
        """
            execute the conversion for the instance with the first
            rule that provide the conversion (see get_convert_from_to)
        """
        convert_from = instance.__class__.__name__
        if convert_from == "Contract":
            convert_from = "ContractContributionPlanDetails"
        list_possible_conversion = cls.get_convert_from_to()
        for possible_conversion in list_possible_conversion:
            if convert_from == possible_conversion['from'] and convert_to == possible_conversion['to']:
                result = cls.convert(instance=instance, convert_to=convert_to, **kwargs)
                return result

    @classmethod
    def get_convert_from_to(cls):
        """
            get the possible conversion, return [calc UUID, from, to]
        """
        list_possible_conversion = []
        for ft in cls.from_to:
            convert_from_to = {'calc_uuid': cls.uuid, 'from': ft['from'], 'to': ft['to']}
            list_possible_conversion.append(convert_from_to)
        return list_possible_conversion


AbsCalculationRule = AbsStrategy



import datetime
import json
import logging
import types
import uuid
from typing import Dict, List, Callable

import graphene
import pandas as pd
from django.db import models
from graphene.types.generic import GenericScalar
from pandas import DataFrame

from core import fields
from core.custom_filters import CustomFilterWizardStorage
from core.models import ExportableQueryModel
from graphql.utils.ast_to_dict import ast_to_dict


logger = logging.getLogger(__file__)


class ExportableQueryMixin:
    export_patches: Dict[str, List[Callable[[DataFrame], DataFrame]]] = {}
    module_name: str
    object_type: str
    related_field: str

    @classmethod
    def get_module_name(cls) -> str:
        if not hasattr(cls, 'module_name') or getattr(cls, 'module_name') is None:
            error_msg = (
                "The class using `ExportableQueryMixin` must specify the `module_name` property "
                "when the customFilters argument is specified. Please override the `module_name` "
                "property in the Query schema to specify how to append custom filters for the specific object."
            )
            raise NotImplementedError(error_msg)
        return cls.module_name

    @classmethod
    def get_object_type(cls) -> str:
        if not hasattr(cls, 'object_type') or getattr(cls, 'object_type') is None:
            error_msg = (
                "The class using `ExportableQueryMixin` must specify the `object_type` property "
                "when the customFilters argument is specified. Please override the `object_type` "
                "property in the Query schema to specify how to append custom filters for the specific object."
            )
            raise NotImplementedError(error_msg)
        return cls.object_type

    @classmethod
    def get_related_field(cls):
        if not hasattr(cls, 'related_field') or getattr(cls, 'related_field') is None:
            return None
        return cls.related_field

    @classmethod
    def get_exportable_fields(cls):
        if not hasattr(cls, 'exportable_fields'):
            raise NotImplementedError(
                "Class using `ExportableQueryMixin` has to provide either exportable_fields "
                "or overwrite`get_exportable_fields` to provide list of fields that can be exported.")
        return cls.exportable_fields

    @classmethod
    def get_patches_for_field(cls, field_type):
        if field_type not in cls.get_exportable_fields():
            raise AttributeError(f"Field {field_type} is not being exported.")
        return cls.export_patches.get(field_type)

    @classmethod
    def __init_subclass_with_meta__(cls, **meta_options):
        cls._setup_exportable_fields(**meta_options)
        return super(ExportableQueryMixin, cls).__init_subclass_with_meta__(**meta_options)

    @classmethod
    def _setup_exportable_fields(cls, **meta_options):
        for field in cls.get_exportable_fields():
            field_name = F"{field}_export"
            fld = getattr(cls, field)
            new_args = fld.args
            new_args['file_format'] = graphene.String()
            new_args['fields'] = graphene.List(of_type=graphene.String)
            new_args['fields_columns'] = GenericScalar()
            setattr(cls, field_name, graphene.Field(graphene.String, args=new_args))
            cls.create_export_function(field)

    @classmethod
    def _adjust_notation(cls, field):
        from graphene.utils.str_converters import to_snake_case, to_camel_case
        return to_snake_case(field.replace('.', '__'))

    @classmethod
    def create_export_function(cls, field_name):
        new_function_name = f"resolve_{field_name}_export"
        default_resolve = getattr(cls, F"resolve_{field_name}", None)

        if not default_resolve:
            raise AttributeError(
                f"Query {cls} doesn't provide resolve function for {field_name}. "
                f"CSV export cannot be created")

        def exporter(cls, self, info, **kwargs):
            custom_filters = kwargs.pop("customFilters", None)
            export_fields = [cls._adjust_notation(f) for f in kwargs.pop('fields')]
            fields_mapping = json.loads(kwargs.pop('fields_columns'))
            file_format = kwargs.pop('file_format', 'csv')

            source_field = getattr(cls, field_name)
            filter_kwargs = {k: v for k, v in kwargs.items() if k in source_field.filtering_args}

            qs = default_resolve(None, info, **kwargs)
            qs = qs.filter(**filter_kwargs)
            qs = cls.__append_custom_filters(custom_filters, qs)
            export_file = ExportableQueryModel\
                .create_csv_export(qs, export_fields, info.context.user, column_names=fields_mapping,
                                   patches=cls.get_patches_for_field(field_name), file_format=file_format)

            return export_file.name

        setattr(cls, new_function_name, types.MethodType(exporter, cls))

    @classmethod
    def __append_custom_filters(cls, custom_filters, queryset):
        if custom_filters:
            module_name = cls.get_module_name()
            object_type = cls.get_object_type()
            related_field = cls.get_related_field()
            queryset = CustomFilterWizardStorage.build_custom_filters_queryset(
                module_name,
                object_type,
                custom_filters,
                queryset,
                relation=related_field
            )
        return queryset



# Generated by Django 3.0.14 on 2022-01-04 11:04
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0015_missing_roles"),
    ]

    operations = [
        migrations.RunSQL(
            "ALTER TABLE tblUsers ADD LastLogin [datetime] NULL"
            if settings.MSSQL else
            'ALTER TABLE "tblUsers" ADD "LastLogin" timestamp NULL',
            reverse_sql='ALTER TABLE "tblUsers" DROP COLUMN "LastLogin"',
        ),
    ]



# Generated by Django 4.2.15 on 2024-09-03 10:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0029_alter_interactiveuser_role_id'),
    ]

    operations = [
        migrations.AddField(
            model_name='exportablequerymodel',
            name='file_format',
            field=models.CharField(blank=True, choices=[('csv', 'csv'), ('xlsx', 'xlsx')], default='csv', max_length=255, null=True),
        ),
    ]



from django.db import connection, migrations
import os


class RemoveIndexForField(migrations.RunPython):
    def __init__(self, app_name, model_name, field_name) -> None:
        self.app_name = app_name
        self.model_name = model_name
        self.field_name = field_name
        super().__init__(self.remove_index, self.reverse_remove_index)

    def remove_index(self, app, schema_editor):
        model = app.get_model(self.app_name, self.model_name)
        field = model._meta.get_field(self.field_name)

        with connection.cursor() as cursor:
            constraints = connection.introspection.get_constraints(cursor, model._meta.db_table)
            for constraint_name, constraint_info in constraints.items():
                if constraint_info["index"] and any(
                        field.column.lower() == col.lower() for col in constraint_info["columns"]):
                    if os.environ.get("DB_DEFAULT") == 'mssql':
                        cursor.execute(f"DROP INDEX {constraint_name} ON {model._meta.db_table}")
                    else:
                        cursor.execute(f'DROP INDEX "{constraint_name}"')

    def reverse_remove_index(self, app, schema_editor):
        pass



from core.models.openimis_graphql_test_case import openIMISGraphQLTestCase, BaseTestContext
from core.test_helpers import create_test_interactive_user
from core import filter_validity
from location.models import Location
import json


class gqlTest(openIMISGraphQLTestCase):
    admin_user = None
    admin_username = "Adminlogin"
    admin_password = "EdfmD3!12@#"
    district = None
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.admin_user = create_test_interactive_user(username=cls.admin_username, password=cls.admin_password)
        cls.admin_token = BaseTestContext(user=cls.admin_user).get_jwt()
        cls.disctict = Location.objects.filter(type='D', *filter_validity()).first()
    
    def test_login_successful(self):
        variables = {
            "username": str(self.admin_username),
            "password": str(self.admin_password)
        }

        query = """
            mutation authenticate($username: String!, $password: String!) {
                tokenAuth(username: $username, password: $password)
                {
                refreshExpiresIn
                }
            }
        """
        response = self.query(
            query,
            variables=variables
        )
        self.assertResponseNoErrors(response)
        _ = json.loads(response.content)

    def test_login_default_successful(self):
        variables = {
            "username": "Admin",
            "password": "admin123"
        }

        query = """
            mutation authenticate($username: String!, $password: String!) {
                tokenAuth(username: $username, password: $password)
                {
                refreshExpiresIn
                }
            }
        """
        response = self.query(
            query,
            variables=variables
        )
        self.assertResponseNoErrors(response)
        _ = json.loads(response.content)

    def test_login_wrong_credentials(self):
        variables = {
            "username": str(self.admin_username),
            "password": "notright"
        }

        query = """
            mutation authenticate($username: String!, $password: String!) {
                tokenAuth(username: $username, password: $password)
                {
                refreshExpiresIn
                }
            }
        """
        response = self.query(
            query,
            variables=variables
        )
        self.assertResponseHasErrors(response)
        _ = json.loads(response.content)

    def test_change_langue(self):
        query = """
            mutation {
            changeUserLanguage(
                input: {clientMutationId: "b2a639a9-1a85-4643-bf84-69d05160c8ee",
                clientMutationLabel: "Change User Language",
                languageId: "fr"}
            ) {
                clientMutationId
                internalId
            }
            }
        """
        response = self.send_mutation_raw(query, self.admin_token)


    def test_create_role(self):
    
        input_param =  {
            "name": "SP Enrollment Officer",
            "isBlocked": False,
            "isSystem": False,
            "rightsId": [159001,159002,159003,159004,159005,180001,180002,180003,180004]
        }

        response = self.send_mutation('createRole',input_param, self.admin_token)
    def test_create_user_with_null_uuid(self):
        query = """
            mutation (
                $input: CreateUserMutationInput!
            ){     
                createUser(input: $input) 
                {   
                    clientMutationId       
                    internalId
                }
            }
        """
        variables = {
            "input": {
                "uuid": None,
                "username": "OCM-176",
                "userTypes": [
                    "INTERACTIVE"
                ],
                "lastName": "add",
                "otherNames": "user",
                "email": "ocm-176@openimis.org",
                "password": "pOCM-176!OCM-176",
                "healthFacilityId": None,
                "districts": [
                    self.disctict.id
                ],
                "locationId": None,
                "language": "en",
                "roles": [
                    "4"
                ],
                "substitutionOfficerId": None,
                "clientMutationLabel": "Create user",
                "clientMutationId": "95b431f3-0c12-40ad-bc01-51034702366d"
            }
        }
        response = self.query(
            query,
            variables=variables,
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"}
        )
        self.assertResponseNoErrors(response)
        self.get_mutation_result("95b431f3-0c12-40ad-bc01-51034702366d", self.admin_token)
    
    def test_user_district_query(self):
        query = """
    {
      userDistricts
      {
        id,uuid,code,name,parent{id, uuid, code, name}
      }
    }
    """
        response = self.query(
            query,
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"}
        )
        self.assertResponseNoErrors(response)
        
        
    def test_fetch_claimadmin(self):
        query = """
      query ClaimAdminPicker ($search: String, $hf: String, $region_uuid: String, $district_uuid: String) {
          claimAdmins(search: $search, first: 20, healthFacility_Uuid: $hf, regionUuid: $region_uuid, districtUuid: $district_uuid) {
              edges {
                  node {
                      id
                      uuid
                      code
                      lastName
                      otherNames
                      healthFacility {
                          id uuid code name level
                          servicesPricelist{id, uuid}, itemsPricelist{id, uuid}
                          location {
                              id
                              uuid
                              code
                              name
                              parent {
                                code name id uuid
                              }
                          }
                      }
                      
                    }
                }
            }
        }
        """
        response = self.query(
            query,
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.admin_token}"}
        )
        self.assertResponseNoErrors(response)
        


import logging

from core.data_masking.masking_class_registry_point import MaskingClassRegistryPoint

logger = logging.getLogger(__name__)


class MaskingClassStorage:

    @classmethod
    def get_masking_class(cls, masking_model: str):
        masking_class = None
        for method_info in MaskingClassRegistryPoint.REGISTERED_MASKING_CLASS:
            if method_info['model'] == masking_model:
                masking_class = method_info['class_reference']
        return masking_class



from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _


class UniqueCodeValidationMixin:
    CODE_DUPLICATE_MSG = _("Object code %(code)s is not unique.")

    @classmethod
    def validate_unique_code_name(cls, code, excluded_id=None, code_key='code'):
        if not cls._unique_code_name(code, excluded_id, code_key=code_key):
            raise ValidationError(cls.CODE_DUPLICATE_MSG % {'code': code})

    @classmethod
    def _unique_code_name(cls, code, excluded_id=None, code_key='code'):
        query = cls.OBJECT_TYPE.objects.filter(**{code_key: code})
        if excluded_id:
            query = query.exclude(id=excluded_id)

        return not query.exists()



# Generated by Django 3.0.3 on 2021-03-23 12:24

import core.fields
import datetime
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0009_mutationlog_client_mutation_details'),
    ]

    operations = [
        migrations.CreateModel(
            name='RoleMutation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('validity_from', core.fields.DateTimeField(db_column='ValidityFrom', default=datetime.datetime.now)),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('legacy_id', models.IntegerField(blank=True, db_column='LegacyID', null=True)),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='roles', to='core.MutationLog')),
                ('role', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to='core.Role')),
            ],
            options={
                'db_table': 'core_RoleMutation',
                'managed': True,
            },
        ),
    ]



# Generated by Django 2.1.11 on 2019-10-04 14:21

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0005_group_usergroup'),
    ]

    operations = [
        migrations.CreateModel(
            name='FieldControl',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('field', models.CharField(max_length=250, unique=True)),
                ('usage', models.IntegerField(blank=True, max_length=1, null=True)),
                ('module', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='controls', to='core.ModuleConfiguration')),
            ],
            options={
                'db_table': 'core_FieldControl',
            },
        ),
    ]



# Generated by Django 3.0.3 on 2021-05-26 21:59

import core.models
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0011_auto_20210324_1528'),
    ]
    from django.apps import apps
    try:
        ClaimAdmin = apps.get_model('claim', 'ClaimAdmin')
    except:
        ClaimAdmin = None
        
    if ClaimAdmin:
        dependencies.append(('claim', '0012_item_service_jsonExtField'))
        
    operations = [
        migrations.AddField(
            model_name='user',
            name='claim_admin',
            #TOREVERT
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='claim.ClaimAdmin' if ClaimAdmin else 'core.ClaimAdmin'),
        ),
        migrations.AddField(
            model_name='user',
            name='officer',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='core.Officer'),
        ),
        migrations.CreateModel(
            name='UserMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='users', to='core.MutationLog')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'core_UserMutation',
                'managed': True,
            },
            bases=(models.Model, core.models.ObjectMutation),
        ),
    ]



# Generated by Django 4.2.20 on 2025-04-04 10:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0030_exportablequerymodel_file_format"),
    ]

    operations = [
        migrations.AlterField(
            model_name="mutationlog",
            name="client_mutation_id",
            field=models.CharField(
                blank=True, db_index=True, max_length=255, null=True
            ),
        ),
    ]



from core.reports import user_activity, registers_status

report_definitions = [
    {
        "name": "user_activity",
        "engine": 0,
        "default_report": user_activity.template,
        "description": "User activity",
        "module": "core",
        "python_query": user_activity.user_activity_query,
        "permission": ["131207"],
    },
    {
        "name": "registers_status",
        "engine": 0,
        "default_report": registers_status.template,
        "description": "Registers status",
        "module": "core",
        "python_query": registers_status.registers_status_query,
        "permission": ["131209"],
    },
]



import os
import datetime
import uuid
import decimal
from django.test import TestCase
from django.db import connections
from django.test.runner import DiscoverRunner
from django.test.utils import get_unique_databases_and_mirrors

from core.utils import full_class_name, comparable, to_json_safe_value
from core.datetimes.ad_datetime import AdDate, AdDatetime


class ComparableTest(TestCase):
    def test_generic_eq(self):
        @comparable
        class A(object):
            def __init__(self, f):
                self.f = f

            def __eq__(self, other):
                return self.f == other.f

        @comparable
        class B(object):
            def __init__(self, f):
                self.f = f

            def __eq__(self, other):
                return self.f == other.f

        obj1 = A(f='a')
        obj2 = A(f='a')
        self.assertEquals(obj1, obj2)
        obj3 = B(f='b')
        self.assertNotEquals(obj1, obj3)
        obj4 = B(f='a')
        self.assertNotEquals(obj1, obj4)


class UtilsTestCase(TestCase):
    def test_full_class_name(self):
        self.assertEquals(full_class_name(
            self), 'core.tests.test_utils.UtilsTestCase')

        self.assertEquals(full_class_name(
            1), 'int')

    def test_json_serialize_value(self):
        self.assertEquals(to_json_safe_value(42), 42)
        self.assertEquals(to_json_safe_value("foo"), "foo")

        uuid_obj = uuid.uuid4()
        self.assertEquals(to_json_safe_value(uuid_obj), str(uuid_obj))

        date_obj = datetime.date(2025, 1, 1)
        self.assertEquals(to_json_safe_value(date_obj), str(date_obj))

        ad_date_obj = AdDate(2025, 1, 1)
        self.assertEquals(to_json_safe_value(ad_date_obj), str(ad_date_obj))

        ad_datetime_obj = AdDatetime(2025, 1, 1, 12, 0, 0)
        self.assertEquals(to_json_safe_value(ad_datetime_obj), str(ad_datetime_obj))

        decimal_obj = decimal.Decimal("12345.6789")
        self.assertEquals(to_json_safe_value(decimal_obj), str(decimal_obj))



# Generated by Django 3.2.16 on 2024-01-24 11:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0025_mutationlog_json_ext'),
    ]

    operations = [
        migrations.AddField(
            model_name='mutationlog',
            name='autogenerated_code',
            field=models.TextField(blank=True, null=True),
        ),
    ]



from django.test import TestCase
from core.models import User, TechnicalUser, InteractiveUser


class UserTestCase(TestCase):

    def test_t_user_active_status(self):
        always_valid = User(username='always_valid',
                            t_user=TechnicalUser(username='always_valid'))
        self.assertTrue(always_valid.is_active)

        from core import datetime, datetimedelta
        not_yet_active = User(username='not_yet_active',
                              t_user=TechnicalUser(username='not_yet_active',
                                                   validity_from=datetime.datetime.now() + datetimedelta(days=1)))
        self.assertFalse(not_yet_active.is_active)

        not_active_anymore = User(username='not_active_anymore',
                                  t_user=TechnicalUser(username='not_active_anymore',
                                                       validity_to=datetime.datetime.now() + datetimedelta(days=-1)))
        self.assertFalse(not_active_anymore.is_active)

    def test_i_active_status(self):
        always_valid = User(username='always_valid',
                            i_user=InteractiveUser(login_name='always_valid'))
        self.assertTrue(always_valid.is_active)

        from core import datetime, datetimedelta
        not_yet_active = User(username='always_valid',
                              i_user=InteractiveUser(login_name='not_yet_active',
                                                     validity_from=datetime.datetime.now() + datetimedelta(days=1)))
        self.assertFalse(not_yet_active.is_active)

        not_active_anymore = User(username='always_valid',
                                  i_user=InteractiveUser(login_name='not_active_anymore',
                                                         validity_to=datetime.datetime.now() + datetimedelta(days=-1)))
        self.assertFalse(not_active_anymore.is_active)



from abc import ABC
from typing import Type
import asyncio
from django.db import transaction

from core.models import HistoryModel, MutationLog
from core.services.utils import check_authentication as check_authentication, output_exception, \
    model_representation, output_result_success, build_delete_instance_payload
from core.validation.base import BaseModelValidation
from core.utils import to_json_safe_value


class BaseService(ABC):

    @property
    def OBJECT_TYPE(self) -> Type[HistoryModel]:
        """
        Django ORM model. It's expected that it'll be inheriting from HistoryModel.
        """
        raise NotImplementedError("Class has to define OBJECT_TYPE for service.")

    def __init__(self, user, validation_class: Type[BaseModelValidation] = BaseModelValidation):
        self.user = user
        self.validation_class = validation_class

    @check_authentication
    def create(self, obj_data):
        try:
            with transaction.atomic():
                obj_data = self._adjust_create_payload(obj_data)
                self.validation_class.validate_create(self.user, **obj_data)
                obj_ = self.OBJECT_TYPE(**obj_data)
                return self.save_instance(obj_)
        except Exception as exc:
            return output_exception(model_name=self.OBJECT_TYPE.__name__, method="create", exception=exc)

    @check_authentication
    def update(self, obj_data):
        try:
            with transaction.atomic():
                obj_data = self._adjust_update_payload(obj_data)
                self.validation_class.validate_update(self.user, **obj_data)
                obj_ = self.OBJECT_TYPE.objects.filter(id=obj_data['id']).first()
                obj_.update(data=obj_data, user=self.user, save=False)
                return self.save_instance(obj_)
        except Exception as exc:
            return output_exception(model_name=self.OBJECT_TYPE.__name__, method="update", exception=exc)

    @check_authentication
    def delete(self, obj_data):
        try:
            with transaction.atomic():
                self.validation_class.validate_delete(self.user, **obj_data)
                obj_ = self.OBJECT_TYPE.objects.filter(id=obj_data['id']).first()
                return self.delete_instance(obj_)
        except Exception as exc:
            return output_exception(model_name=self.OBJECT_TYPE.__name__, method="delete", exception=exc)

    def save_instance(self, obj_):
        obj_.save(user=self.user, username=self.user.username)
        dict_repr = model_representation(obj_)
        return output_result_success(dict_representation=dict_repr)

    def delete_instance(self, obj_):
        obj_.delete(user=self.user, username=self.user.username)
        return build_delete_instance_payload()

    def _adjust_create_payload(self, payload_data):
        return self._base_payload_adjust(payload_data)

    def _adjust_update_payload(self, payload_data):
        self._align_json_ext(payload_data)
        return self._base_payload_adjust(payload_data)

    def _align_json_ext(self, payload_data):
        json_ext = payload_data.get('json_ext')
        if isinstance(json_ext, dict):
            for key, value in payload_data.items():
                if key in json_ext and json_ext[key] != value:
                    json_ext[key] = to_json_safe_value(value)

    def _base_payload_adjust(self, obj_data):
        return obj_data


def wait_for_mutation(client_mutation_id):
    mutation = MutationLog.objects.filter(
        client_mutation_id=client_mutation_id
    ).first()
    if not mutation:
        return
    loop_count = 0
    while mutation.status == MutationLog.RECEIVED and loop_count<10:
        asyncio.sleep(0.3)
        loop_count += 1
    return



# Generated by Django 3.2.20 on 2023-08-31 10:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0024_alter_usergroup_options'),
    ]

    operations = [
        migrations.AddField(
            model_name='mutationlog',
            name='json_ext',
            field=models.JSONField(blank=True, db_column='JsonExt', null=True),
        ),
    ]



import logging
from functools import lru_cache
from core.utils import insert_role_right_for_system
from django.db import migrations, models

logger = logging.getLogger(__name__)

ROLE_RIGHTS_ID = [101201, 101101, 101001, 101105]
MANAGER_ROLE_IS_SYSTEM = 64  # By default missing rights are assigned to ClaimAdmin role



def create_role_right(apps, schema_editor):
    RoleRight = apps.get_model('core', 'RoleRight')
    Role = apps.get_model('core', 'Role')
    for right_id in ROLE_RIGHTS_ID:
        insert_role_right_for_system(256, right_id, apps )


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0014_users')
    ]

    operations = [
        migrations.AddField(
            model_name='roleright',
            name='role',
            field=models.ForeignKey(db_column='RoleID', on_delete=models.deletion.DO_NOTHING, related_name='rights', to='core.role'),
        ),
        migrations.RunPython(create_role_right),
    ]



from core import views
from django.urls import include, path
from rest_framework.routers import DefaultRouter

router = DefaultRouter()
router.register(r'users', views.UserViewSet)

urlpatterns = [
    path('', include(router.urls)),
    path('fetch_export', views.fetch_export),
    path('scheduled_jobs', views.get_scheduled_jobs)
]



# Generated by Django 2.1.11 on 2019-10-08 09:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0006_fieldcontrol'),
    ]

    operations = [
        migrations.AlterField(
            model_name='fieldcontrol',
            name='usage',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



from .jwt import jwt_decode_user_key

from rest_framework.authentication import BaseAuthentication
from rest_framework.exceptions import Throttled
from rest_framework import exceptions
from graphql_jwt.utils import get_credentials
from graphql_jwt.exceptions import JSONWebTokenError
from graphql_jwt.shortcuts import get_user_by_token
from core.apps import CoreConfig
from django.conf import settings
from django_ratelimit.core import is_ratelimited

from datetime import date
import jwt
import logging

logger = logging.getLogger(__file__)


class JWTAuthentication(BaseAuthentication):
    """
    class to obtain token from header if it is provided
    and verify if this is correct/valid token
    """

    def authenticate(self, request):
        self.check_rate_limit(request)
        token = get_credentials(request)
        if token:
            # Do not pass context to avoid to try to get user from request to get his private key.
            try:
                user = get_user_by_token(token)
            except (jwt.PyJWTError, JSONWebTokenError) as exc:
                raise exceptions.AuthenticationFailed("INCORRECT_CREDENTIALS") from exc
            except Exception as exc:
                raise exceptions.AuthenticationFailed(str(exc)) from exc
            else:
                if CoreConfig.is_valid_health_facility_contract_required:
                    if not (hasattr(user, 'health_facility') and hasattr(user.health_facility, 'contract_end_date') and
                            user.health_facility.contract_end_date > date.today()):
                        raise exceptions.AuthenticationFailed("HF_CONTRACT_INVALID")

            return user, None

    def enforce_csrf(self, request):
        return  # To not perform the csrf during checking auth header

    @staticmethod
    def check_rate_limit(request) -> None:
        group = settings.RATELIMIT_GROUP
        key = settings.RATELIMIT_KEY
        rate = settings.RATELIMIT_RATE
        mode = settings.MODE

        if mode == 'PROD' and is_ratelimited(
                request=request,
                group=group,
                fn=None,
                key=key,
                rate=rate,
                method=is_ratelimited.ALL,
                increment=True
        ):
            raise Throttled(detail='Rate limit exceeded')



from core.models import Officer, InteractiveUser, User, TechnicalUser, filter_validity
from core.models.openimis_graphql_test_case import openIMISGraphQLTestCase
from core.services.userServices import create_or_update_officer_villages, create_or_update_interactive_user, \
    create_or_update_core_user
from core.services import create_or_update_user_roles
from location.models import Location
from uuid import uuid4


def create_test_officer(valid=True, custom_props=None, villages=[]):
    if custom_props is None:
        custom_props = {}
    else:
        custom_props = {k: v for k, v in custom_props.items() if hasattr(Officer, k)}

    code = custom_props.pop('code', None)
    uuid = custom_props.pop('uuid', None)
    qs_eo = Officer.objects
    eo = None
    data = {
        "code": code or "TSTOFF",
        "uuid": uuid,
        "last_name": "Officer",
        "other_names": "Test",
        "validity_to": None if valid else "2019-06-01",
        "audit_user_id": -1,
        "phone": "0000110100",
        **custom_props
    }

    if code:
        qs_eo = qs_eo.filter(code=code)
    if uuid:
        qs_eo = qs_eo.filter(uuid=uuid)
    eo = None
    if code or uuid:
        eo = qs_eo.first()
    if eo:
        data['uuid'] = eo.uuid
        eo.update(data)
    else:
        data['uuid'] = uuid4()
        eo = Officer.objects.create(**data)
    
    if not villages:
        villages == Location.objects.filter(*filter_validity(), type='V').first()
    if eo:
        _ = create_or_update_officer_villages(eo, [v.id for v in villages], 1)
        return eo


def create_test_interactive_user(username='TestInteractiveTest', password="S\\:\\/pe®Pąßw0rd""", roles=None,
                                 custom_props=None):
    if custom_props is None:
        custom_props = {}
    else:
        custom_props = {k: v for k, v in custom_props.items() if hasattr(InteractiveUser, k)}
    if roles is None:
        roles = [7, 1, 2, 3, 4, 5, 6]
    user = None
    i_user = InteractiveUser.objects.filter(login_name=username).first()
    if i_user:
        # TODO add custom prop to existing user
        user = User.objects.filter(i_user=i_user).first()
    else:
        i_user = InteractiveUser.objects.create(
            **{
                "language_id": "en",
                "last_name": "TestLastName",
                "other_names": "Test Other Names",
                "login_name": username,
                "audit_user_id": -1,
                "role_id": roles[0],
                **custom_props
            }
        )
        
    if not user:
        user = User.objects.create(
            username=username,
            i_user=i_user,
        )
    i_user.set_password(password)
    i_user.save()
    create_or_update_user_roles(i_user, roles, None)
    return user


def create_test_technical_user(
        username='TestAdminTechnicalTest', password="S\\/pe®Pąßw0rd""", super_user=False,
        custom_tech_user_props={}, custom_core_user_props={}):
    custom_tech_user_props['password'] = password
    t_user, t_user_created = TechnicalUser.objects.get_or_create(
        **{
            "username": username,
            "email": "test_tech_user@openimis.org",
            "is_staff": super_user,
            "is_superuser": super_user,
            **(custom_tech_user_props)
        }
    )
    # Just for safety and retrieving the User because TechnicalUser will automatically create its User
    custom_core_user_props['password'] = password
    core_user, core_user_created = User.objects.get_or_create(
        username=username,
        t_user=t_user,
        **(custom_core_user_props)
    )
    return core_user


def compare_dicts(dict1, dict2):
    def recursive_compare(obj1, obj2):
        if isinstance(obj1, dict) and isinstance(obj2, dict):
            # Check keys
            if set(obj1.keys()) != set(obj2.keys()):
                return False

            # Recursively compare values
            for key in obj1.keys():
                if not recursive_compare(obj1[key], obj2[key]):
                    return False

            return True
        elif isinstance(obj1, list) and isinstance(obj2, list):
            # Check list length
            if len(obj1) != len(obj2):
                return False

            # Recursively compare list elements
            for item1, item2 in zip(obj1, obj2):
                if not recursive_compare(item1, item2):
                    return False

            return True
        elif isinstance(obj1, (float, int)) or (isinstance(obj1, str) and obj1.isnumeric()) \
                and isinstance(obj2, (float, int)) or (isinstance(obj2, str) and obj2.isnumeric()):
            # Compare floating-point numbers with a tolerance for decimal precision
            return round(float(obj1), 2) == round(float(obj2), 2)

        # Compare other types directly
        return obj1 == obj2

    return recursive_compare(dict1, dict2)


def AssertMutation(test_obj, mutation_uuid, token):
    return openIMISGraphQLTestCase().get_mutation_result(mutation_uuid, token)


class LogInHelper:
    def __init__(self):
        self.test_user_name = "TestUserTest2"
        self.test_user_password = "TestPasswordTest2@"
        self.test_data_user = {
            "username": self.test_user_name,
            "last_name": self.test_user_name,
            "password": self.test_user_password,
            "other_names": self.test_user_name,
            "user_types": "INTERACTIVE",
            "language": "en",
            "roles": [1, 3, 5, 9],
        }

    def get_or_create_user_api(self, **kwargs):
        username = kwargs.get('username') or self.test_user_name
        user = User.objects.filter(username=username).first()
        if user is None:
            user = self._create_user_interactive_core(**kwargs)
        return user

    def _create_user_interactive_core(self, **kwargs):
        username = kwargs.get('username') or self.test_user_name
        i_user, i_user_created = create_or_update_interactive_user(
            user_id=None, data={**self.test_data_user, **kwargs}, audit_user_id=999, connected=False)
        create_or_update_core_user(
            user_uuid=None, username=username, i_user=i_user)
        return User.objects.get(username=username)



import graphene
from core.schema import OpenIMISMutation


class DeleteInputType(OpenIMISMutation.Input):
    uuids = graphene.List(graphene.UUID)


class ReplaceInputType(OpenIMISMutation.Input):
    uuid = graphene.UUID(required=True)


import json
from typing import Union

from django.contrib.auth.models import AnonymousUser
from django.contrib.contenttypes.models import ContentType
from django.core.serializers.json import DjangoJSONEncoder
from django.forms.models import model_to_dict


def check_authentication(function):
    def wrapper(self, *args, **kwargs):
        if type(self.user) is AnonymousUser or not self.user.id:
            return {
                "success": False,
                "message": "Authentication required",
                "detail": "PermissionDenied",
            }
        else:
            result = function(self, *args, **kwargs)
            return result

    return wrapper


def check_permissions(permissions=None):
    def decorator(function):
        def wrapper(self, *args, **kwargs):
            if not self.user.has_perms(permissions):
                return {
                    "success": False,
                    "message": "Permissions required",
                    "detail": "PermissionDenied",
                }
            else:
                result = function(self, *args, **kwargs)
                return result

        return wrapper

    return decorator


def model_representation(model):
    uuid_string = str(model.id)
    dict_representation = model_to_dict(model)
    dict_representation["id"], dict_representation["uuid"] = (str(uuid_string), str(uuid_string))
    return dict_representation


def output_exception(model_name, method, exception):
    return {
        "success": False,
        "message": f"Failed to {method} {model_name}",
        "detail": str(exception),
        "data": "",
    }


def output_result_success(dict_representation):
    return {
        "success": True,
        "message": "Ok",
        "detail": "",
        "data": json.loads(json.dumps(dict_representation, cls=DjangoJSONEncoder)),
    }


def build_delete_instance_payload():
    return {
        "success": True,
        "message": "Ok",
        "detail": "",
    }


def get_generic_type(generic_type: Union[str, ContentType]):
    if isinstance(generic_type, ContentType):
        return generic_type
    return ContentType.objects.get(model__iexact=str(generic_type).lower())



import sys
import datetime as py_datetime
from functools import total_ordering
from .shared import datetimedelta

__all__ = ["tzinfo", "timezone", "AdDate", "date", "AdDatetime", "datetime"]

"""
Standard Gregorian date,
wrapped with openIMIS data handling helpers
"""

core = sys.modules["core"]

tzinfo = py_datetime.tzinfo
timezone = py_datetime.timezone

@total_ordering
class AdDate(py_datetime.date):

    @classmethod
    def today(cls):
        return AdDate.from_ad_date(py_datetime.date.today())

    @classmethod
    def from_ad_date(cls, value):
        if value is None:
            return None
        return AdDate(value.year, value.month, value.day)

    def to_ad_date(self):
        return self

    @classmethod
    def from_ad_datetime(cls, value):
        return AdDate.from_ad_date(value)

    def to_ad_datetime(self):
        return AdDatetime(self.year, self.month, self.day)

    def to_datetime(self):
        return self.to_ad_datetime()

    def raw_isoformat(self):
        return self.isoformat()

    def ad_isoformat(self):
        return self.isoformat()

    def displayshortformat(self):
        return self.strftime(core.shortstrfdate)

    def displaylongformat(self):
        return self.strftime(core.longstrfdate)

    @classmethod
    def _convert_op_res(cls, res):
        if isinstance(res, py_datetime.datetime):
            return AdDate.from_ad_datetime(res)
        if isinstance(res, py_datetime.date):
            return AdDate.from_ad_date(res)
        return res

    def __add__(self, other):
        if isinstance(other, datetimedelta):
            return datetimedelta.add_to_date(other, self)
        dt = super(AdDate, self).__add__(other)
        return AdDate._convert_op_res(dt)

    def __sub__(self, other):
        if isinstance(other, datetimedelta):
            return datetimedelta.add_to_date((other * -1), self)
        dt = super(AdDate, self).__sub__(other)
        return AdDate._convert_op_res(dt)

    def __repr__(self):
        L = [self.year, self.month, self.day]
        if L[-1] == 0:
            del L[-1]
        if L[-1] == 0:
            del L[-1]
        return "%s.date(%s)" % (self.__class__.__module__, ", ".join(map(str, L)))
    
    def _date_operation(self, operation, other):

        if not other:
            return operation(other)
        if isinstance(other, py_datetime.date):
            return operation(other)
        if isinstance(other, py_datetime.datetime):
            return operation(AdDatetime.from_ad_datetime(other))

    def __eq__(self, other):
        result = self._date_operation(super(AdDate, self).__eq__, other)
        return result if result else self - other == datetimedelta()

    def __gt__(self, other):
        return self._date_operation(super(AdDate, self).__gt__, other)

    def __lt__(self, other):
        return self._date_operation(super(AdDate, self).__lt__, other)

    
date = AdDate
date.min = AdDate(1, 1, 1)
date.max = AdDate(9999, 12, 31)

@total_ordering
class AdDatetime(py_datetime.datetime):

    @classmethod
    def now(cls):
        return cls.from_ad_datetime(py_datetime.datetime.now())

    @classmethod
    def from_ad_date(cls, value):
        if value is None:
            return None
        return AdDatetime(value.year, value.month, value.day)

    def to_ad_date(self):
        return AdDate(self.year, self.month, self.day)

    @classmethod
    def from_ad_datetime(cls, value):
        if value is None:
            return None
        return AdDatetime(value.year, value.month, value.day,
                          value.hour, value.minute, value.second, value.microsecond,
                          value.tzinfo)

    def __hash__(self):
        return super().__hash__()

    def to_ad_datetime(self):
        return self

    def _date_operation(self, operation, other):

        if not other:
            return operation(other)
        if isinstance(other, py_datetime.datetime):
            return operation(other)
        if isinstance(other, py_datetime.date):
            return operation(AdDatetime.from_ad_date(other))

    def __eq__(self, other):
        result = self._date_operation(super(AdDatetime, self).__eq__, other)
        return result if result else self - other == datetimedelta()

    def __gt__(self, other):
        return self._date_operation(super(AdDatetime, self).__gt__, other)

    def __lt__(self, other):
        return self._date_operation(super(AdDatetime, self).__lt__, other)

    @classmethod
    def _convert_op_res(cls, res):
        if isinstance(res, py_datetime.datetime):
            return AdDatetime.from_ad_datetime(res)
        return res

    def __add__(self, other):
        if isinstance(other, datetimedelta):
            return datetimedelta.add_to_date(other, self)
        dt = super(AdDatetime, self).__add__(other)
        return AdDatetime._convert_op_res(dt)

    def __sub__(self, other):
        if isinstance(other, datetimedelta):
            return datetimedelta.add_to_date((other * -1), self)
        dt = super(AdDatetime, self).__sub__(other)
        return AdDatetime._convert_op_res(dt)

    def __repr__(self):
        L = [self.year, self.month, self.day,
             self.hour, self.minute, self.second, self.microsecond]
        if L[-1] == 0:
            del L[-1]
        if L[-1] == 0:
            del L[-1]
        s = "%s.datetime(%s)" % (self.__class__.__module__,
                                 ", ".join(map(str, L)))
        if self.tzinfo is not None:
            assert s[-1:] == ")"
            s = s[:-1] + ", tzinfo=%r" % self.tzinfo + ")"
        if self.fold:
            assert s[-1:] == ")"
            s = s[:-1] + ", fold=1)"
        return s


datetime = AdDatetime



from django.test import TestCase
from core.test_helpers import create_test_officer, LogInHelper
import re
from core.models import InteractiveUser, Officer, User


class HelpersTest(TestCase):

    def test_create_test_officer(self):
        count_before = Officer.objects.count()
        user = create_test_officer(valid=True, custom_props={})
        self.assertTrue(type(user) is Officer)
        count_after = Officer.objects.count()
        self.assertEquals(count_after, count_before + 1)

    def test_login_helper(self):
        count_before = InteractiveUser.objects.count()
        login_helper = LogInHelper()
        user = login_helper.get_or_create_user_api()
        self.assertTrue(type(user) is User)
        count_after = InteractiveUser.objects.count()
        self.assertEquals(count_after, count_before + 1)


class GQLTest(TestCase):
    _DATA_MUTATION_VARS = {
        "variables": {
            "input": {
                "birthDate": "1990-10-18",
                "clientMutationId": "6a7ff306-24d0-4b51-a383-c072e764e842",
                "clientMutationLabel": "Update user",
                "districts": [
                    "20"
                ],
                "email": "Admin@admin.com",
                "healthFacilityId": "6",
                "language": "en",
                "lastName": "Manal",
                "locationId": None,
                "otherNames": "Roger",
                "phone": None,
                "roles": [
                    "9"
                ],
                "substitutionOfficerId": None,
                "username": "VHOS0011",
                "userTypes": [
                    "INTERACTIVE",
                    "CLAIM_ADMIN"
                ],
                "uuid": "28026414-f7dd-48fa-bedd-2f4cce9f9677"
            }
        }
    }

    def test_save(self):

        _ = self.to_camel_case_key(self._DATA_MUTATION_VARS['variables']['input'])
        # create_or_update_interactive_user(self.user,data)
        # self.user=InteractiveUser.objects.filter(uuid = '28026414-f7dd-48fa-bedd-2f4cce9f9677').first()
        # self.assertEquals(self.user.lastName, "Manal")

    def to_camel_case_key(self, input):
        pattern = re.compile(r'(?<!^)(?=[A-Z]|[0-9]+)')
        if isinstance(input, list):
            res = []
            for elm in input:
                res.append(self.to_camel_case_key(elm))
            return res
        elif isinstance(input, dict):
            res = {}
            for key in input:
                res[pattern.sub('_', key).lower()] = self.to_camel_case_key(input[key])
            return res
        else:
            return input



import importlib
import core
from django.test import TestCase
from datetime import date as py_date


class CalendarTestCase(TestCase):
    def setUp(self):
        super(CalendarTestCase, self).setUp()
        core.calendar = importlib.import_module(
            '.calendars.ne_calendar', 'core')
        core.datetime = importlib.import_module(
            '.datetimes.ne_datetime', 'core')

    def tearDown(self):
        core.calendar = importlib.import_module(
            '.calendars.ad_calendar', 'core')
        core.datetime = importlib.import_module(
            '.datetimes.ad_datetime', 'core')

    def test_from_ad_date(self):
        dt = core.datetime.date.from_ad_date(py_date(2020, 1, 13))
        self.assertEqual(dt, core.datetime.date(2076, 9, 28))

    def test_weekfirstday(self):
        dt = core.datetime.date(2076, 9, 28)
        fwd = core.calendar.weekfirstday(dt)
        self.assertEqual(fwd, core.datetime.date(2076, 9, 27))

    def test_weeklastday(self):
        dt = core.datetime.date(2076, 9, 28)
        fwd = core.calendar.weeklastday(dt)
        self.assertEqual(fwd, core.datetime.date(2076, 10, 4))

    def test_weekday(self):
        wd = core.calendar.weekday(2076, 9, 11)
        self.assertEqual(wd, 5)
        wd = core.calendar.weekday(2076, 9, 28)
        self.assertEqual(wd, 1)

    def test_monthfirstday(self):
        dt = core.calendar.monthfirstday(2076, 9)
        self.assertEqual(dt, core.datetime.date(2076, 9, 1))

    def test_monthlastday(self):
        dt = core.calendar.monthlastday(2019, 2)
        self.assertEqual(dt, core.datetime.date(2019, 2, 32))

    def test_yearfirstday(self):
        dt = core.calendar.yearfirstday(2075)
        self.assertEqual(dt, core.datetime.date(2075, 1, 1))

    def test_yearlastday(self):
        dt = core.calendar.yearlastday(2075)
        self.assertEqual(dt, core.datetime.date(2075, 12, 30))

    def test_monthdayscount(self):
        self.assertEqual(32, core.calendar.monthdayscount(2019, 2))
        self.assertEqual(29, core.calendar.monthdayscount(2075, 10))

    def test_yeardayscount(self):
        self.assertEqual(365, core.calendar.yeardayscount(2076))
        self.assertEqual(366, core.calendar.yeardayscount(2077))



import logging
from django.contrib.auth import logout
from django.contrib.sessions.models import Session
from django.utils.timezone import now
from django_ratelimit.core import is_ratelimited
from django.http import HttpResponse
from rest_framework.exceptions import JsonResponse
from django.conf import settings


logger = logging.getLogger(__name__)


class DefaultAxesAttributesMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Set default values for Django-axes attributes if they're not already set
        if not hasattr(request, 'axes_ip_address'):
            request.axes_ip_address = request.META.get('REMOTE_ADDR', '')
        if not hasattr(request, 'axes_user_agent'):
            request.axes_user_agent = request.META.get('HTTP_USER_AGENT', '')
        if not hasattr(request, 'axes_attempt_time'):
            request.axes_attempt_time = now()

        return self.get_response(request)


class SecurityHeadersMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)

        if settings.MODE == "PROD":
            response["Strict-Transport-Security"] = "max-age=63072000; includeSubDomains"
            response["Content-Security-Policy"] = "default-src 'self';"
            response["X-Frame-Options"] = "DENY"
            response["X-Content-Type-Options"] = "nosniff"
            response["Referrer-Policy"] = "strict-origin-when-cross-origin"
            response["Permissions-Policy"] = "geolocation=(), microphone=()"

        return response


class GraphQLRateLimitMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        group = settings.RATELIMIT_GROUP
        key = settings.RATELIMIT_KEY
        rate = settings.RATELIMIT_RATE
        mode = settings.MODE
        if mode == 'PROD' and request.path == '/api/graphql':
            rate_limited = is_ratelimited(
                request=request,
                group=group,
                key=key,
                rate=rate,
                method=is_ratelimited.ALL,
                increment=True
            )
            if rate_limited:
                return JsonResponse({'detail': 'Rate limit exceeded'}, status=429)
        response = self.get_response(request)
        return response


class AdminLogoutMiddleware:
    """
    Middleware to clear all user sessions when they log out from Django Admin.
    """
    LOGOUT_URL = f"/{settings.SITE_ROOT()}admin/logout/"

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)

        if request.path.startswith(self.LOGOUT_URL):
            response.delete_cookie('JWT')
            logger.info(f"Cleared all sessions after admin panel logout")

        return response



# Generated by Django 3.2.19 on 2023-06-01 11:45
from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):

    psql_code = 'select 1'
    mssql_code = 'ALTER TABLE [tblOfficer] ALTER COLUMN [JsonExt]'

    dependencies = [
        ('core', '0022_auto_20230516_1742'),
    ]

    operations = [
        migrations.RunSQL(f'{mssql_code} NVARCHAR(MAX)'
                          if settings.MSSQL else psql_code,
                          reverse_sql=f'{mssql_code} TEXT'
                          if settings.MSSQL else psql_code),
    ]



# Generated by Django 3.2.20 on 2023-07-14 08:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0023_alter_jsonext_column_in_tblOfficer'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='usergroup',
            options={'managed': False},
        ),
    ]



import uuid
from copy import copy
from datetime import datetime as py_datetime

from django.db import models

from core.utils import (
    CachedManager,
    CachedModelMixin
    
)
from ..fields import DateTimeField
from ..utils import filter_validity
import logging

logger = logging.getLogger(__name__)



class BaseVersionedModel(CachedModelMixin, models.Model):
    validity_from = DateTimeField(db_column='ValidityFrom', default=py_datetime.now)
    validity_to = DateTimeField(db_column='ValidityTo', blank=True, null=True)

    # Use our custom CachedManager for object retrieval
    objects = CachedManager()

    def update(self, *args, **kwargs):
        """
        Overrides the default update to update the cache after saving the instance.
        """
        obj_data = kwargs.pop('data', {})
        if not obj_data:
            obj_data = kwargs
            kwargs = {}
        [setattr(self, key, obj_data[key]) for key in obj_data]
        self.save(*args, **kwargs)

    def save(self, *args, **kwargs):
        """
        Overrides the default save to update the cache after saving the instance.
        """
        caching = kwargs.pop('cache_update', True)
        super().save(*args, **kwargs)
        if caching:
            # Build the cache key using the same logic as in the CachedManager.
            # (Assuming lookups are done using pk/id/uuid)
            self.update_cache()
        else:
            self.delete_cache()
        return self

    def delete(self, *args, **kwargs):
        """
        Overrides the default delete to remove the instance from the cache.
        """
        # Build the cache key prior to deletion.
        self.delete_cache()
        # Then perform the actual deletion.
        return super().delete(*args, **kwargs)

    def save_history(self, **kwargs):
        if not self.id:  # only copy if the data is being updated
            return None
        histo = copy(self)
        histo.id = None
        if hasattr(histo, "uuid"):
            setattr(histo, "uuid", uuid.uuid4())
        histo.validity_to = py_datetime.now()
        histo.legacy_id = self.id
        histo.save(cache_update=False)
        return histo.id

    def delete_history(self, **kwargs):
        self.save_history()
        now = py_datetime.now()
        self.validity_from = now
        self.validity_to = now
        self.save()

    class Meta:
        abstract = True

    @classmethod
    def filter_queryset(cls, queryset=None):
        if queryset is None:
            queryset = cls.objects.all()
        queryset = queryset.filter(*filter_validity())
        return queryset




class VersionedModel(BaseVersionedModel):
    legacy_id = models.IntegerField(
        db_column='LegacyID', blank=True, null=True)



    class Meta:
        abstract = True


class UUIDVersionedModel(BaseVersionedModel):
    legacy_id = models.UUIDField(
        db_column='LegacyID', blank=True, null=True)

    class Meta:
        abstract = True





import sys
import os
import importlib
import logging
from django.apps import AppConfig
from django.conf import settings

logger = logging.getLogger(__name__)

MODULE_NAME = "core"

this = sys.modules[MODULE_NAME]

DEFAULT_CFG = {
    "username_code_length": "12",  # cannot be bigger than 50 unless modified length limit
    "username_changeable": True,
    "auto_provisioning_user_group": "user",
    "calendar_package": "core",
    "calendar_module": ".calendars.ad_calendar",
    "datetime_package": "core",
    "datetime_module": ".datetimes.ad_datetime",
    "shortstrfdate": "%d/%m/%Y",
    "longstrfdate": "%a %d %B %Y",
    "iso_raw_date": "False",
    "age_of_majority": "18",
    "async_mutations": "True" if os.environ.get("ASYNC", os.environ.get("MODE", "PROD")) == "PROD" else "False",
    "password_reset_template": "password_reset.txt",
    "currency": "$",
    "gql_query_claim_admins_perms" : [],
    "gql_query_users_perms": ["121701"],
    "gql_mutation_create_users_perms": ["121702"],
    "gql_mutation_update_users_perms": ["121703"],
    "gql_mutation_delete_users_perms": ["121704"],
    "gql_query_roles_perms": ["122001"],
    "gql_mutation_create_roles_perms": ["122002"],
    "gql_mutation_update_roles_perms": ["122003"],
    "gql_mutation_replace_roles_perms": ["122006"],
    "gql_mutation_duplicate_roles_perms": ["122005"],
    "gql_mutation_delete_roles_perms": ["122004"],
    # TODO consider moving that roles related to ClaimAdmin and EnrolmentOfficer
    #  into modules related to that type of user for example
    #  EnrolmentOfficer -> policy module, ClaimAdmin -> claim module etc
    "gql_query_enrolment_officers_perms": ["121501"],
    "gql_mutation_create_enrolment_officers_perms": ["121502"],
    "gql_mutation_update_enrolment_officers_perms": ["121503"],
    "gql_mutation_delete_enrolment_officers_perms": ["121504"],
    "gql_query_claim_administrator_perms": ["121601"],
    "gql_mutation_create_claim_administrator_perms": ["121602"],
    "gql_mutation_update_claim_administrator_perms": ["121603"],
    "gql_mutation_delete_claim_administrator_perms": ["121604"],
    "fields_controls_user": {},
    "fields_controls_eo": {},
    "is_valid_health_facility_contract_required": False,
    "secondary_calendar": None,
    "locked_user_password_hash": 'locked',
    "gql_query_enable_viewing_masked_data_perms": ["900101"],
    "csrf_protect_login": True,
}


class CoreConfig(AppConfig):
    default_auto_field = 'django.db.models.AutoField'  # Django 3.1+
    name = MODULE_NAME
    username_code_length = 12
    username_changeable = True
    age_of_majority = 18
    password_reset_template = "password_reset.txt"
    
    gql_query_claim_admins_perms = []
    gql_query_roles_perms = []
    gql_mutation_create_roles_perms = []
    gql_mutation_update_roles_perms = []
    gql_mutation_replace_roles_perms = []
    gql_mutation_duplicate_roles_perms = []
    gql_mutation_delete_roles_perms = []
    gql_query_users_perms = []
    gql_mutation_create_users_perms = []
    gql_mutation_update_users_perms = []
    gql_mutation_delete_users_perms = []
    # TODO consider moving that roles related to ClaimAdmin and EnrolmentOfficer
    #  into modules related to that type of user for example
    #  EnrolmentOfficer -> policy module, ClaimAdmin -> claim module etc
    gql_query_enrolment_officers_perms = []
    gql_mutation_create_enrolment_officers_perms = []
    gql_mutation_update_enrolment_officers_perms = []
    gql_mutation_delete_enrolment_officers_perms = []
    gql_query_claim_administrator_perms = []
    gql_mutation_create_claim_administrator_perms = []
    gql_mutation_update_claim_administrator_perms = []
    gql_mutation_delete_claim_administrator_perms = []
    is_valid_health_facility_contract_required = None
    locked_user_password_hash = None

    fields_controls_user = {}
    fields_controls_eo = {}
    secondary_calendar = None

    password_min_length = settings.PASSWORD_MIN_LENGTH
    password_uppercase = settings.PASSWORD_UPPERCASE
    password_lowercase = settings.PASSWORD_LOWERCASE
    password_digits = settings.PASSWORD_DIGITS
    password_symbols = settings.PASSWORD_SYMBOLS

    gql_query_enable_viewing_masked_data_perms = []

    csrf_protect_login = None

    def _import_module(self, cfg, k):
        logger.info('import %s.%s' %
                    (cfg["%s_module" % k], cfg["%s_package" % k]))
        return importlib.import_module(
            cfg["%s_module" % k], package=cfg["%s_package" % k])

    def _configure_calendar(self, cfg):
        this.shortstrfdate = cfg["shortstrfdate"]
        this.longstrfdate = cfg["longstrfdate"]
        this.iso_raw_date = False if cfg["iso_raw_date"] is None else cfg["iso_raw_date"].lower(
        ) == "true"
        try:
            this.calendar = self._import_module(cfg, "calendar")
            this.datetime = self._import_module(cfg, "datetime")
        except Exception:
            logger.error('Failed to configure calendar, using default!\n%s: %s' % (
                sys.exc_info()[0].__name__, sys.exc_info()[1]))
            this.calendar = self._import_module(DEFAULT_CFG, "calendar")
            this.datetime = self._import_module(DEFAULT_CFG, "datetime")

    def _configure_user_config(self, cfg):
        this.username_code_length = int(cfg["username_code_length"])
        # Quick fix, this config has to be rebuilt
        CoreConfig.username_code_length = int(cfg["username_code_length"])
        CoreConfig.username_changeable = cfg["username_changeable"]

    def _configure_majority(self, cfg):
        this.age_of_majority = int(cfg["age_of_majority"])

    def _configure_currency(self, cfg):
        this.currency = str(cfg["currency"])

    def _configure_auto_provisioning(self, cfg):
        if bool(os.environ.get('NO_DATABASE', False)):
            logger.info('env NO_DATABASE set to True: no user auto provisioning possible!')
            return
        group = cfg["auto_provisioning_user_group"]
        this.auto_provisioning_user_group = group
        try:
            from .models import Group
            Group.objects.get(name=group)
        except Group.DoesNotExist:
            g = Group(name=group)
            g.save()
            from django.contrib.auth.models import Permission
            p = Permission.objects.get(codename="view_user")
            g.permissions.add(p)
            g.save()
        except Exception as e:
            logger.warning('Failed set auto_provisioning_user_group ' + str(e))

    def _configure_graphql(self, cfg):
        this.async_mutations = True if cfg["async_mutations"] is None else cfg["async_mutations"].lower() == "true"

    def _configure_permissions(self, cfg):
        CoreConfig.gql_query_claim_admins_perms = cfg["gql_query_claim_admins_perms"]
        CoreConfig.gql_query_roles_perms = cfg["gql_query_roles_perms"]
        CoreConfig.gql_mutation_create_roles_perms = cfg["gql_mutation_create_roles_perms"]
        CoreConfig.gql_mutation_update_roles_perms = cfg["gql_mutation_update_roles_perms"]
        CoreConfig.gql_mutation_replace_roles_perms = cfg["gql_mutation_replace_roles_perms"]
        CoreConfig.gql_mutation_duplicate_roles_perms = cfg["gql_mutation_duplicate_roles_perms"]
        CoreConfig.gql_mutation_delete_roles_perms = cfg["gql_mutation_delete_roles_perms"]
        CoreConfig.gql_query_users_perms = cfg["gql_query_users_perms"]
        CoreConfig.gql_mutation_create_users_perms = cfg["gql_mutation_create_users_perms"]
        CoreConfig.gql_mutation_update_users_perms = cfg["gql_mutation_update_users_perms"]
        CoreConfig.gql_mutation_delete_users_perms = cfg["gql_mutation_delete_users_perms"]
        CoreConfig.gql_query_enrolment_officers_perms = cfg["gql_query_enrolment_officers_perms"]
        CoreConfig.gql_mutation_create_enrolment_officers_perms = cfg["gql_mutation_create_enrolment_officers_perms"]
        CoreConfig.gql_mutation_update_enrolment_officers_perms = cfg["gql_mutation_update_enrolment_officers_perms"]
        CoreConfig.gql_mutation_delete_enrolment_officers_perms = cfg["gql_mutation_delete_enrolment_officers_perms"]
        CoreConfig.gql_query_claim_administrator_perms = cfg["gql_query_claim_administrator_perms"]
        CoreConfig.gql_mutation_create_claim_administrator_perms = cfg["gql_mutation_create_claim_administrator_perms"]
        CoreConfig.gql_mutation_update_claim_administrator_perms = cfg["gql_mutation_update_claim_administrator_perms"]
        CoreConfig.gql_mutation_delete_claim_administrator_perms = cfg["gql_mutation_delete_claim_administrator_perms"]
        CoreConfig.gql_mutation_delete_claim_administrator_perms = cfg["gql_mutation_delete_claim_administrator_perms"]
        CoreConfig.gql_query_enable_viewing_masked_data_perms = cfg["gql_query_enable_viewing_masked_data_perms"]
        CoreConfig.csrf_protect_login = cfg["csrf_protect_login"]

        CoreConfig.fields_controls_user = cfg["fields_controls_user"]
        CoreConfig.fields_controls_eo = cfg["fields_controls_eo"]

    def _configure_additional_settings(self, cfg):
        CoreConfig.is_valid_health_facility_contract_required = cfg["is_valid_health_facility_contract_required"]
        CoreConfig.secondary_calendar = cfg["secondary_calendar"]

    def ready(self):
        from .models import ModuleConfiguration
        cfg = ModuleConfiguration.get_or_default(MODULE_NAME, DEFAULT_CFG)
        self._configure_calendar(cfg)
        self._configure_user_config(cfg)
        self._configure_majority(cfg)
        self._configure_auto_provisioning(cfg)
        self._configure_graphql(cfg)
        self._configure_currency(cfg)
        self._configure_permissions(cfg)
        self._configure_additional_settings(cfg)

        CoreConfig.password_reset_template = cfg["password_reset_template"]
        CoreConfig.locked_user_password_hash = cfg["locked_user_password_hash"]

        # The scheduler starts as soon as it gets a job, which could be before Django is ready, so we enable it here
        from core import scheduler
        if settings.SCHEDULER_AUTOSTART:
            scheduler.start()



import json

from core.websocket.abstract_websocket_client import AbstractWebSocketClient


class BaseWebSocketClient(AbstractWebSocketClient):
    def receive(self):
        """
        If no receive actions were determined, this method returns all messages received from socket.
        """
        return self.__messages

    def _default_action(self, message):
        self.__messages.append(message)

    def _transform_payload(self, payload):
        return str(payload)


class JsonWebSocketClient(BaseWebSocketClient):

    def _transform_payload(self, payload):
        payload = json.dumps(payload)
        return payload



from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('core', '0018_auto_20230318_1551'),
    ]

    operations = [
        migrations.RunSQL('ALTER TABLE [tblOfficer] ADD [JsonExt] TEXT'
                          if settings.MSSQL else
                          'ALTER TABLE "tblOfficer" ADD "JsonExt" jsonb'),

    ]


import json
from django.conf import settings
from graphene_django.utils.testing import GraphQLTestCase
import uuid
import logging
from graphene import Schema
from graphene.test import Client
import datetime
import time
from django.test import RequestFactory
from django.middleware.csrf import get_token
from graphql_jwt.shortcuts import get_token as get_token_jwt
from rest_framework.request import Request as DRFRequest
from django.contrib.auth.models import AnonymousUser
from django.contrib.sessions.backends.db import SessionStore

logger = logging.getLogger(__name__)

class BaseTestContext:
    def __init__(self, user=None, method="GET", path="/", data=None, headers=None):
        """
        Initialize a test context with realistic request attributes.
        
        Args:
            user: User instance (authenticated or None for anonymous).
            method: HTTP method (e.g., "GET", "POST").
            path: URL path (e.g., "/api/endpoint/").
            data: Request payload (dict for POST/PUT, None for GET).
            headers: Custom HTTP headers (dict).
        """
        cookies = {}
        self.factory = RequestFactory()
        
        # Initialize session
        self.session = SessionStore()
        if user is not None:
            self.user = user
            self.session.create()  # Create a new session
            self.session['user_id'] = str(user.id)  # Store user ID or other relevant data
            self.session.save()  # Save session to generate session_key
            self.jwt = get_token_jwt(self.user, self)
            cookies['JWT'] = self.jwt 
        else:
            self.user = AnonymousUser()

        # Create request
        self.method = method.upper()
        self.request = self.factory.generic(
            method=self.method,
            path=path,
            data=json.dumps(data) if data else {},  # Ensure JSON data is serialized
            content_type="application/json"
        )
        
        # Attach session and user to request
        self.request.session = self.session
        self.request.user = self.user
        
        # Set up META dictionary
        self.META = self.request.META
        self.META["REQUEST_METHOD"] = self.method
        self.META["PATH_INFO"] = path
        self.META["SERVER_NAME"] = "testserver"
        self.META["SERVER_PORT"] = "80"

        # Add CSRF token if needed
        if self.method in ["POST", "PUT", "PATCH"]:
            self.META["CSRF_COOKIE"] = get_token(self.request)
            self.request.CSRF_TOKEN = self.META["CSRF_COOKIE"]

        # Add CORS headers
        self.META["HTTP_ORIGIN"] = "http://testclient.com"
        self.META["HTTP_ACCESS_CONTROL_REQUEST_METHOD"] = self.method

        # Add custom headers
        if headers:
            for key, value in headers.items():
                meta_key = f"HTTP_{key.upper().replace('-', '_')}"
                self.META[meta_key] = value

        # Add cookies (e.g., session ID and JWT token)
        cookies['sessionid'] = self.session.session_key
        if cookies:
            cookie_string = "; ".join(f"{key}={value}" for key, value in cookies.items())
            self.META["HTTP_COOKIE"] = cookie_string

    def update_meta(self, key, value):
        """Utility method to update META dictionary."""
        self.META[key] = value
        self.request.META = self.META

    def get_request(self):
        """Return the constructed request object."""
        return self.request

    def get_jwt(self):
        """Return the JWT token."""
        return getattr(self, 'jwt', None)
        
        

class openIMISGraphQLTestCase(GraphQLTestCase):
    GRAPHQL_URL = f"/{settings.SITE_ROOT()}graphql"
    GRAPHQL_SCHEMA = True

    
    # client = None
    @classmethod
    def setUpClass(cls):
        # cls.client=Client(cls.schema)
        super(openIMISGraphQLTestCase, cls).setUpClass()

    def get_mutation_result(self, mutation_uuid, token, internal=False, allow_exceptions=True):
        content = None
        while True:
            # wait for the mutation to be done
            # wait for the mutation to be done
            if internal:
                filter_uuid = f""" id: "{mutation_uuid}" """
            else:
                filter_uuid = f""" clientMutationId: "{mutation_uuid}" """

            response = self.query(
                f"""
                {{
                mutationLogs({filter_uuid})
                {{
                pageInfo {{ hasNextPage, hasPreviousPage, startCursor, endCursor}}
                edges
                {{
                    node
                    {{
                        id,status,error,{'clientMutationId,' if not internal else ''}clientMutationLabel,clientMutationDetails,requestDateTime,jsonExt
                    }}
                }}
                }}
                }}

                """,
                headers={"HTTP_AUTHORIZATION": f"Bearer {token}"},
            )
            content = json.loads(response.content)
            if "data" in content:
                if "mutationLogs" in content["data"]:
                    if "edges" in content["data"]["mutationLogs"]:
                        for e in content["data"]["mutationLogs"]["edges"]:
                            if "node" in e:
                                e = e["node"]
                                if e and "status" in e and e["status"] != 0:
                                    self._assert_mutationEdge_no_error(e)
                                    return content
                else:
                    if allow_exceptions:
                        raise ValueError("mutation has no edge field")
                    else:
                        logger.error("mutation has no edge field")
                        return content
            else:
                if allow_exceptions:
                    raise ValueError("mutation has no data field")
                else:
                    logger.error("mutation has no data field")
                    return content
            time.sleep(1)
        if self._assert_mutationEdge_no_error(content):
            return None

    def _assert_mutationEdge_no_error(self, e):

        if "error" in e and e["error"]:
            raise ValueError(
                f"At least one edge of the mutation has error: {e['error']}"
            )
            return False
        elif "errors" in e and e["errors"]:
            raise ValueError(
                f"At least one edge of the mutation has error: {e['errors']}"
            )
            return False
        elif "status" in e and e["status"] == 1:
            raise ValueError("Mutation failed with status 1")
            return False
        return True

    def send_mutation_raw(self, mutation_raw, token, variables_param=None, follow=True):
        params = {"headers": {"HTTP_AUTHORIZATION": f"Bearer {token}"}}
        if variables_param:
            params["variables"] = variables_param
        response = self.query(
            mutation_raw,
            **params,
        )
        self.assertResponseNoErrors(response)
        content = json.loads(response.content)

        if follow:
            mutation_type = list(content['data'].keys())[0]
            return self.get_mutation_result(
                content['data'][mutation_type]['internalId'],
                token,
                internal=True
            )
        else:
            return json.loads(response.content)

    def send_mutation(self, mutation_type, input_params, token, follow=True, raw=False, add_client_mutation_id=False, allow_exceptions=True):
        # copy to avoid adding clientMutationId to the calling param
        input_params = dict(input_params)
        if add_client_mutation_id and "clientMutationId" not in input_params:
            input_params["clientMutationId"] = str(uuid.uuid4())
        response = self.query(
            f"""
        mutation 
        {{
            {mutation_type}(input: {{
               {input_params if raw else self.build_params(input_params)}
            }})  

          {{
            internalId
            {'clientMutationId' if 'clientMutationId' in input_params else ''} 
          }}
        }}
        """,
            headers={"HTTP_AUTHORIZATION": f"Bearer {token}"},
        )
        if allow_exceptions:
            self.assertResponseNoErrors(response)
        content = json.loads(response.content)
        if follow:
            return self.get_mutation_result(
                content["data"][mutation_type]["internalId"], token, internal=True, allow_exceptions=allow_exceptions
            )
        else:
            return content

    # This validates the status code and if you get errors
    def build_params(self, params):
        def wrap_arg(v):
            if isinstance(v, str):
                return f'"{v}"'
            if isinstance(v, list):
                return f"[{','.join([str(wrap_arg(vv)) for vv in v])}]"
            if isinstance(v, dict):
                return str(self.build_params(v))
            if isinstance(v, bool):
                return str(v).lower()
            if isinstance(v, datetime.date):
                return graphene.DateTime.serialize(
                    datetime.datetime.fromordinal(v.toordinal())
                )
            return v

        params_as_args = [
            f"{k}:{wrap_arg(v)}" for k, v in params.items() if v is not None
        ]
        return ", ".join(params_as_args)



import queue

from enum import Enum

from django import dispatch
from typing import Callable


class ServiceSignalBindType(Enum):
    BEFORE = 0
    AFTER = 1
    BEFORE_AND_AFTER = 2


class RegisteredServiceSignal:

    def __init__(self, providing_args=None):
        """
        Due to how python manage class prototypes, there's a chance that class definition using register_service_signal
        will not be loaded until first import of given class. Therefore, it's likely that binding function responsible
        for loading functions called before/after signal will be invoked before signal register. In this case functions
        are queued and connected to actual signal after the registration. Whether registration took place
        is defined by providing_args. If argument is None, signal is considered not registered.
        """
        self.__signal_results = {'before': None, 'after': None}
        self.__connection_queue = {type_: queue.Queue() for type_ in ServiceSignalBindType}
        self.__signals_before = []
        self.__signals_after = []

        if not providing_args:
            self.__before_service_signal = None
            self.__after_service_signal = None
        else:
            self.__before_service_signal = dispatch.Signal(providing_args)
            self.__after_service_signal = dispatch.Signal(providing_args + ['result'])

    @property
    def connected_signals(self):
        return {
            'before': self.__signals_before,
            'after': self.__signals_after
        }

    def send_signal_before(self, sender, **signal_call_args):
        result = self.before_service_signal.send(sender=sender, **signal_call_args)
        self.__signal_results['before'] = result

    def send_signal_after(self, sender, **signal_call_args):
        result = self.after_service_signal.send(sender=sender, **signal_call_args)
        self.__signal_results['after'] = result

    @property
    def signal_results(self):
        return self.__signal_results

    @property
    def before_service_signal(self):
        return self.__before_service_signal

    @property
    def after_service_signal(self):
        return self.__after_service_signal

    def connect_signal(self, func: Callable, bind_type: ServiceSignalBindType = ServiceSignalBindType.BEFORE_AND_AFTER):
        if self.is_signal_registered():
            self._add_connection(func, bind_type)
        else:
            self.__connection_queue[bind_type].put(func)

    def register_signal(self, providing_args):
        if self.is_signal_registered():
            raise ValueError(f"Signal {self} already registered.")

        self.__before_service_signal = dispatch.Signal(providing_args)
        self.__after_service_signal = dispatch.Signal(providing_args + ['result'])
        self._connect_queued()

    def is_signal_registered(self):
        return self.__before_service_signal is not None and self.__after_service_signal is not None

    def _add_connection(self, func, bind_type):
        if bind_type == ServiceSignalBindType.BEFORE:
            self.__before_service_signal.connect(func)
            self.__signals_before.append(func)
        elif bind_type == ServiceSignalBindType.AFTER:
            self.__after_service_signal.connect(func)
            self.__signals_after.append(func)
        elif bind_type == ServiceSignalBindType.BEFORE_AND_AFTER:
            self.__before_service_signal.connect(func)
            self.__after_service_signal.connect(func)
            self.__signals_after.append(func)
            self.__signals_before.append(func)
        else:
            raise AttributeError(f"Invalid bind_type {bind_type}, should be one of {ServiceSignalBindType}.")

    def _connect_queued(self):
        for binding_type, awaiting_queue in self.__connection_queue.items():
            while not awaiting_queue.empty():
                self.connect_signal(awaiting_queue.get(), binding_type)
